#! /usr/bin/env python
import os
import tempfile
import urllib
import zipfile

import yaml
from model_metadata.model_data_files import FileTemplate
from model_metadata.model_setup import FileSystemLoader
from model_metadata import ModelMetadata

from ..utils import as_cwd


def _parse_author_info(info):
    """Get list of authors as a tuple."""
    author = info.get("author", info.get("authors", ""))
    if isinstance(author, str):
        author = (author,)
    return tuple(author)


class SetupMixIn(object):
    def __init__(self):
        name = self.__class__.__name__.split(".")[-1]

        self._meta = ModelMetadata.from_obj(self._bmi)

        self._defaults = {}
        self._parameters = {}
        for name, param in self._meta.parameters.items():
            self._parameters[name] = param["value"]["default"]
            try:
                units = param["units"]
            except KeyError:
                units = None
            self._defaults[name] = param["value"], units

    def setup(self, *args, **kwds):
        """Set up a simulation.

        Parameters
        ----------
        path : str, optional
            Path to a folder to set up the simulation. If not given,
            use a temporary folder.

        Returns
        -------
        str
            Path to the folder that contains the set up simulation.
        """
        if len(args) == 0:
            dir_ = tempfile.mkdtemp()
        else:
            dir_ = args[0]

        self._parameters.update(kwds)

        FileSystemLoader(self.datadir).stage_all(dir_, **self._parameters)

        config = self._meta.run["config_file"]
        if config["contents"] and not config["path"]:
            config["path"] = dir_

        if config["path"]:
            with as_cwd(dir_):
                config_file = FileTemplate.write(
                    config["contents"], config["path"], **self._parameters
                )
        else:
            config_file = None

        return config_file, os.path.abspath(dir_)

    @property
    def parameters(self):
        return self._parameters.items()

    @property
    def defaults(self):
        return self._defaults.items()

    @property
    def datadir(self):
        return self._meta.base

    @property
    def author(self):
        return _parse_author_info(self._meta.info)

    @property
    def email(self):
        return self._meta.info["email"]

    @property
    def contact(self):
        contact = self.author[0]
        email = self.email
        if email != "-":
            contact = "{name} <{email}>".format(name=contact, email=email)
        return contact

    @property
    def url(self):
        return self._meta.info["url"]

    @property
    def license(self):
        return self._meta.info["license"]

    @property
    def doi(self):
        return self._meta.info["doi"]

    @property
    def version(self):
        return self._meta.info["version"]

    @property
    def summary(self):
        return self._meta.info["summary"]

    @property
    def cite_as(self):
        return self._meta.info["cite_as"]


class GitHubSetupMixIn(object):
    def setup(self, case="default", name=None, path=None):
        if name is None:
            try:
                name = self.name
            except AttributeError:
                name = self.get_component_name()
        input_dir = fetch_input(name, path=path)

        case_dir = os.path.join(input_dir, case)

        return get_initialize_arg(case_dir)


def get_initialize_arg(dir_):
    """Get the BMI initialize argument for a set of input files.

    Parameters
    ----------
    dir_ : str
        Path to a folder that contains input files.

    Returns
    -------
    str
        Argument that can be passed to a BMI initialize method.
    """
    readme = os.path.join(dir_, "README.yaml")
    with open(readme, "r") as fp:
        metadata = yaml.safe_load(fp)

    args = metadata["bmi"]["initialize"]["args"]
    if isinstance(args, dict):
        with tempfile.NamedTemporaryFile("w", dir=dir_, delete=False) as fp:
            fp.write(args["contents"])
            args = fp.name
    return args


def url_to_input_file_repo(name):
    """URL to a repo of input files for a component.

    Parameters
    ----------
    name : str
        Component name.

    Returns
    -------
    str
        URL of input-file repository.
    """
    return "https://github.com/mcflugen/{name}-input".format(name=name)


def url_to_input_file_zip(name):
    """URL of zip file on input files.

    Parameters
    ----------
    name : str
        Component name.

    Returns
    -------
    str
        URL of zip file of input files.
    """
    return os.path.join(url_to_input_file_repo(name), "archive", "master.zip")


def fetch_input(name, path=None):
    """Fetch input for a BMI component.

    Parameters
    ----------
    name : str
        Component name.
    path : str, optional
        Directory to extract into.

    Returns
    -------
    str
        Path to input files for various cases.
    """
    extract_to = os.path.abspath(path or tempfile.mkdtemp())

    (path_, _) = urllib.urlretrieve(url_to_input_file_zip(name))  # nosec

    with zipfile.ZipFile(path_, "r") as zipfp:
        zipfp.extractall(path=extract_to)

    return os.path.join(extract_to, "{name}-input-master".format(name=name))



import numpy as np

from pymt.grids.connectivity import get_connectivity


def raster_node_coordinates(shape, spacing=None, origin=None):
    """Node coordinates of a uniform rectilinear grid.

    Parameters
    ----------
    shape : array_like
        Size of grid in each dimension.
    spacing : array_like (optional)
        Node spacing in each dimension (default is unit spacing).
    origin : array_like (optional)
        Coordinates of the origin node (default is 0.0).

    Returns
    -------
    coords : ndarray
        Node coordinates.

    Examples
    --------
    >>> from pymt.component.grid import raster_node_coordinates
    >>> (y, x) = raster_node_coordinates((2, 3))
    >>> y
    array([[ 0.,  0.,  0.],
           [ 1.,  1.,  1.]])
    >>> x
    array([[ 0.,  1.,  2.],
           [ 0.,  1.,  2.]])

    >>> (x, ) = raster_node_coordinates((3, ), (2., ), (1., ))
    >>> x
    array([ 1.,  3.,  5.])
    """
    spacing = spacing or np.ones_like(shape, dtype=float)
    origin = origin or np.zeros_like(shape, dtype=float)

    coordinate_vectors = []
    for shape_, spacing_, origin_ in zip(shape, spacing, origin):
        start, stop = origin_, origin_ + shape_ * spacing_
        coordinate_vectors.append(np.arange(start, stop, spacing_))

    return np.meshgrid(*coordinate_vectors, indexing="ij")


def get_raster_node_coordinates(grid, grid_id):
    """Get coordinates of nodes on a raster grid.

    Parameters
    ----------
    grid : grid_like
        A raster grid.
    grid_id : int
        Grid identifier.

    Returns
    -------
    coords : ndarray
        Coordinates of the grid nodes.

    Examples
    --------
    >>> from pymt.component.grid import get_raster_node_coordinates
    >>> class RasterGrid(object):
    ...     def get_grid_shape(self, grid_id):
    ...         return (2, 3)
    ...     def get_grid_spacing(self, grid_id):
    ...         return (1., 2.)
    ...     def get_grid_origin(self, grid_id):
    ...         return (2., 1.)
    >>> g = RasterGrid()
    >>> (y, x) = get_raster_node_coordinates(g, 0)
    >>> y
    array([[ 2.,  2.,  2.],
           [ 3.,  3.,  3.]])
    >>> x
    array([[ 1.,  3.,  5.],
           [ 1.,  3.,  5.]])
    """
    (shape, spacing, origin) = (
        grid.get_grid_shape(grid_id),
        grid.get_grid_spacing(grid_id),
        grid.get_grid_origin(grid_id),
    )

    return raster_node_coordinates(shape, spacing=spacing, origin=origin)


def get_rectilinear_node_coordinates(grid, grid_id):
    """Get coordinates of nodes on a rectilinear grid.

    Parameters
    ----------
    grid: grid_like
        A rectilinear grid.
    grid_id : int
        Grid identifier.

    Returns
    -------
    coords : ndarray
        Coordinates of the grid nodes.

    Examples
    --------
    >>> from pymt.component.grid import get_rectilinear_node_coordinates
    >>> class RectilinearGrid(object):
    ...     def get_grid_x(self, grid_id):
    ...         return (0., 3., 4)
    ...     def get_grid_y(self, grid_id):
    ...         return (2., 7.)
    >>> g = RectilinearGrid()
    >>> (y, x) = get_rectilinear_node_coordinates(g, 0)
    >>> y
    array([[ 2.,  2.,  2.],
           [ 7.,  7.,  7.]])
    >>> x
    array([[ 0.,  3.,  4.],
           [ 0.,  3.,  4.]])
    """
    coordinate_vectors = []
    for coordinate_name in ["z", "y", "x"]:
        try:
            func = getattr(grid, "get_grid_" + coordinate_name)
            coordinate_vectors.append(func(grid_id))
        except (AttributeError, NotImplementedError):
            pass

    return np.meshgrid(*coordinate_vectors, indexing="ij")


def get_structured_node_coordinates(grid, grid_id):
    """Get coordinates of nodes on a structured grid.

    Parameters
    ----------
    grid: grid_like
        A structured grid.
    grid_id : int
        Grid identifier.

    Returns
    -------
    coords : ndarray
        Coordinates of the grid nodes.

    Examples
    --------
    >>> from pymt.component.grid import get_structured_node_coordinates
    >>> class StructuredGrid(object):
    ...     def get_grid_x(self, grid_id):
    ...         return np.array((0., 3., 4, 0., 3., 4.)).reshape((2, 3))
    ...     def get_grid_y(self, grid_id):
    ...         return np.array((2., 2., 2., 7., 7., 7.)).reshape((2, 3))
    ...     def get_grid_z(self, grid_id):
    ...         raise NotImplementedError('get_grid_z')
    >>> g = StructuredGrid()
    >>> (y, x) = get_structured_node_coordinates(g, 0)
    >>> y
    array([[ 2.,  2.,  2.],
           [ 7.,  7.,  7.]])
    >>> x
    array([[ 0.,  3.,  4.],
           [ 0.,  3.,  4.]])
    """
    node_coordinates = []
    for coordinate_name in ["z", "y", "x"]:
        try:
            func = getattr(grid, "get_grid_" + coordinate_name)
            node_coordinates.append(np.array(func(grid_id)))
        except (AttributeError, NotImplementedError):
            pass
    return node_coordinates


def get_structured_node_connectivity(grid, grid_id):
    """Get cell connectivity on a structured grid of quadrilaterals.

    Parameters
    ----------
    grid: grid_like
        A structured grid.
    grid_id : int
        Grid identifier.

    Returns
    -------
    conn : ndarray
        Connectivities of nodes to cells.

    Examples
    --------
    >>> from pymt.component.grid import get_structured_node_connectivity
    >>> class StructuredGrid(object):
    ...     def get_grid_shape(self, grid_id):
    ...         return (2, 3)
    >>> g = StructuredGrid()
    >>> (c, o) = get_structured_node_connectivity(g, 0)
    >>> c
    array([0, 1, 4, 3, 1, 2, 5, 4])
    >>> o
    array([4, 8])
    """
    shape = grid.get_grid_shape(grid_id)
    return get_connectivity(shape, with_offsets=True)


class GridMixIn(object):
    """Mix-in that makes a Component grid-like.

    Examples
    --------
    >>> class Port(object):
    ...     def get_component_name(self):
    ...         return 'test-component'
    ...     def get_input_item_count(self):
    ...         return 1
    ...     def get_input_item_list(self):
    ...         return ['invar']
    ...     def get_output_item_count(self):
    ...         return 1
    ...     def get_output_item_list(self):
    ...         return ['outvar']
    ...     def get_grid_shape(self, grid_id):
    ...         return (2, 3)
    ...     def get_grid_spacing(self, grid_id):
    ...         return (2., 1.)
    ...     def get_grid_origin(self, grid_id):
    ...         return (0., 0.)

    >>> from pymt.component.grid import GridMixIn
    >>> class Component(GridMixIn):
    ...     def __init__(self):
    ...         self._port = Port()
    ...         super(Component, self).__init__()
    >>> c = Component()
    >>> c.name
    'test-component'
    >>> c.input_items
    ['invar']
    >>> c.output_items
    ['outvar']
    >>> c.get_grid_type(0)
    'RASTER'
    >>> c.get_x(0)
    array([[ 0.,  1.,  2.],
           [ 0.,  1.,  2.]])
    >>> c.get_y(0)
    array([[ 0.,  0.,  0.],
           [ 2.,  2.,  2.]])
    """

    def __init__(self):
        self._coords = {}
        self._connectivity = {}
        self._grid_type = {}

    @property
    def name(self):
        """Name of the wrapped component."""
        return self._port.get_component_name()

    @property
    def input_items(self):
        """Input item names as a list."""
        if self._port.get_input_item_count() > 0:
            return self._port.get_input_item_list()
        else:
            return []

    @property
    def output_items(self):
        """Output item names as a list."""
        if self._port.get_output_item_count() > 0:
            return self._port.get_output_item_list()
        else:
            return []

    def _set_grid_type(self, grid_id):
        try:
            shape = self._port.get_grid_shape(grid_id)
            if shape is None:
                raise Exception
        except Exception:
            grid_type = "UNSTRUCTURED"
        else:
            try:
                self._port.get_grid_spacing(grid_id)
            except Exception:
                x = self._port.get_grid_x(grid_id)
                if len(x) == shape[-1]:
                    grid_type = "RECTILINEAR"
                else:
                    grid_type = "STRUCTURED"
            else:
                grid_type = "RASTER"
        self._grid_type[grid_id] = grid_type

    def get_grid_type(self, grid_id):
        """The type of the grid.

        Parameters
        ----------
        grid_id : int
            Grid identifier.

        Returns
        -------
        type : str
            The type of the grid.
        """
        try:
            return self._grid_type[grid_id]
        except (AttributeError, KeyError):
            self._set_grid_type(grid_id)
            return self._grid_type[grid_id]

    def _set_coords(self, grid_id):
        grid_type = self.get_grid_type(grid_id)

        if grid_type == "RASTER":
            coords = get_raster_node_coordinates(self._port, grid_id)
        elif grid_type == "RECTILINEAR":
            coords = get_rectilinear_node_coordinates(self._port, grid_id)
        else:
            coords = get_structured_node_coordinates(self._port, grid_id)

        self._coords[grid_id] = coords

    def _get_coord_by_dim(self, grid_id, dim):
        try:
            return self._coords[grid_id][dim]
        except (AttributeError, KeyError):
            self._set_coords(grid_id)
            return self._coords[grid_id][dim]
        except IndexError:
            print(self._port)
            print(grid_id)
            print(self._coords)
            raise

    def _set_connectivity(self, grid_id):
        grid_type = self.get_grid_type(grid_id)
        if grid_type != "UNSTRUCTURED":
            (connectivity, offset) = get_structured_node_connectivity(
                self._port, grid_id
            )
        else:
            (connectivity, offset) = (
                self.get_grid_connectivity(grid_id),
                self.get_grid_offset(grid_id),
            )

        self._connectivity[grid_id] = (connectivity, offset)

    def _get_connectivity(self, grid_id):
        try:
            return self._connectivity[grid_id][0]
        except (AttributeError, KeyError):
            self._set_connectivity(grid_id)
            return self._connectivity[grid_id][0]

    def _get_offset(self, grid_id):
        try:
            return self._connectivity[grid_id][1]
        except (AttributeError, KeyError):
            self._set_connectivity(grid_id)
            return self._connectivity[grid_id][1]

    def get_x(self, grid_id):
        return self._get_coord_by_dim(grid_id, -1)

    def get_y(self, grid_id):
        return self._get_coord_by_dim(grid_id, -2)

    def get_z(self, grid_id):
        return self._get_coord_by_dim(grid_id, -3)

    def get_connectivity(self, grid_id):
        return self._get_connectivity(grid_id)

    def get_offset(self, grid_id):
        return self._get_offset(grid_id)

    def get_value(self, *args, **kwds):
        return self._port.get_value(*args, **kwds)

    def set_value(self, name, values):
        return self._port.set_value(name, values)

    def get_var_units(self, name):
        return self._port.get_var_units(name)



#! /usr/bin/env python

from pymt.grids.grid_type import (
    GridTypeRectilinear,
    GridTypeStructured,
    GridTypeUnstructured,
)

from .constants import open_netcdf
from .ugrid_read import (
    NetcdfRectilinearFieldReader,
    NetcdfStructuredFieldReader,
    NetcdfUnstructuredFieldReader,
)

_NETCDF_MESH_TYPE = {
    "rectilinear": GridTypeRectilinear(),
    "structured": GridTypeStructured(),
    "unstructured": GridTypeUnstructured(),
}

_NETCDF_READERS = {
    "rectilinear": NetcdfRectilinearFieldReader,
    "structured": NetcdfStructuredFieldReader,
    "unstructured": NetcdfUnstructuredFieldReader,
}


def query_netcdf_mesh_type(path, fmt="NETCDF4"):
    root = open_netcdf(path, mode="r", fmt=fmt)

    try:
        type_string = root.variables["mesh"].type
    except AttributeError:
        raise AttributeError("netcdf file is missing type attribute")
    except KeyError:
        raise AttributeError("netcdf file is missing mesh attribute")
    finally:
        root.close()

    try:
        mesh_type = _NETCDF_MESH_TYPE[type_string]
    except KeyError:
        raise TypeError("%s: mesh type not understood" % mesh_type)

    return mesh_type


def field_fromfile(path, fmt="NETCDF4"):
    mesh_type = query_netcdf_mesh_type(path)

    try:
        reader = _NETCDF_READERS[str(mesh_type)]
    except KeyError:
        raise TypeError("%s: no reader available for file" % mesh_type)
    else:
        nc_file = reader(path, fmt=fmt)

    if len(nc_file.times) > 0:
        return (nc_file.fields, nc_file.times)
    else:
        return nc_file.fields



CENTERING_CHOICES = ["zonal", "point"]


class Error(Exception):
    """Base class for grid and field errors"""

    pass


class DimensionError(Error):
    """Error to indicate a dimension mismatch when adding a value field to a grid."""

    def __init__(self, dim0, dim1):
        try:
            self.src_dim = tuple(dim0)
        except TypeError:
            self.src_dim = dim0

        try:
            self.dst_dim = tuple(dim1)
        except TypeError:
            self.dst_dim = dim1

    def __str__(self):
        return "%s != %s" % (self.src_dim, self.dst_dim)


class CenteringValueError(Error):
    """Error to indicate an invalid value for value centering."""

    def __init__(self, val):
        self.val = val

    def __str__(self):
        return "%s: Bad value for 'centering'" % self.val


class GridTypeError(Error):
    def __str__(self):
        try:
            return "Grid is not %s" % self.type
        except AttributeError:
            return "Grid is not of required type"


class NonUniformGridError(GridTypeError):
    """Error to indicate a grid is not a uniform rectilinear grid"""

    type = "uniform rectilinear"


class NonStructuredGridError(GridTypeError):
    """Error to indicate a grid is not a structured grid"""

    type = "structured"


class IGrid(object):
    """An interface for a grid object that represents a structured or unstructured
    grid of nodes and elements.
    """

    def get_x(self):
        """Return the x-coordinates of the grid nodes."""
        pass

    def get_y(self):
        """Return the y-coordinates of the grid nodes."""
        pass

    def get_connectivity(self):
        """Return the connectivity array for the grid."""
        pass

    def get_offset(self):
        """Return an array of offsets in to the connectivity array for each
        element of the grid."""
        pass


class IField(IGrid):
    def get_field(self, field_name):
        """Return an array of values for the requested field"""
        pass



"""Wrap a port as a :class:`Timeline` event.
"""
import os
import sys

import six
import yaml

from ..component.grid import GridMixIn
from ..framework import services
from ..mappers import NearestVal
from ..utils import as_cwd


class PortEvent(GridMixIn):
    """Wrap a port as an event.

    Parameters
    ----------
    port : str
        Name of port.
    init_args : list, optional
        List of arguments to initialize port.
    run_dir : str, optional
        Path to directory to execute port.
    """

    def __init__(self, *args, **kwds):
        if isinstance(kwds["port"], six.string_types):
            self._port = services.get_component_instance(kwds["port"])
        else:
            self._port = kwds["port"]

        self._init_args = kwds.get("init_args", [])
        self._run_dir = kwds.get("run_dir", ".")

        if isinstance(self._init_args, six.string_types):
            self._init_args = [self._init_args]

        self._status_fp = open(
            os.path.abspath(os.path.join(self._run_dir, "_time.txt")), "w"
        )

        GridMixIn.__init__(self)

    def initialize(self):
        """Initialize the event.

        Run the underlying port's initialization method in its *run_dir*. The event's *init_args* are passed to
        the initialize method as arguments.
        """
        with as_cwd(self._run_dir):
            status = {"name": self.name, "time": 0.0, "status": "initializing"}
            self._status_fp.write(yaml.dump(status))
            try:
                self._port.initialize(*self._init_args)
            except Exception:
                raise

    def run(self, time):
        """Run the event.

        Call the `run` method of the underlying port using *time* as the run time.

        Parameters
        ----------
        time : float
            Time to run the event to.
        """
        with as_cwd(self._run_dir):
            status = {"name": self.name, "time": time, "status": "running"}
            print(yaml.dump(status), file=self._status_fp)
            self._status_fp.flush()
            sys.stdout.flush()
            sys.stderr.flush()
            print(yaml.dump(status, default_flow_style=True))

            self._port.run(time)

            sys.stdout.flush()
            sys.stderr.flush()

    def update(self, time):
        with as_cwd(self._run_dir):
            self._port.update_until(time)

    def finalize(self):
        """Finalize the event.

        Run the `finalize` method of the underlying port.
        """
        with as_cwd(self._run_dir):
            status = {"name": self.name, "time": None, "status": "finishing"}
            self._status_fp.write(yaml.dump(status))
            self._port.finalize()
        status = {"name": self.name, "time": None, "status": "finished"}
        self._status_fp.write(yaml.dump(status))
        self._status_fp.close()


class PortMapEvent(object):
    """An event that maps values between ports.

    Parameters
    ----------
    src_port : str
        Port name that is the data source.
    dst_port : str
        Port name that is the destination.
    vars_to_map : list, optional
        Names of variable to map.
    method : {'direct', 'nearest'}, optional
        Method used to map values.
    """

    def __init__(self, *args, **kwds):
        if isinstance(kwds["src_port"], six.string_types):
            self._src = services.get_component_instance(kwds["src_port"])
        else:
            self._src = kwds["src_port"]
        if isinstance(kwds["dst_port"], six.string_types):
            self._dst = services.get_component_instance(kwds["dst_port"])
        else:
            self._dst = kwds["dst_port"]
        self._vars_to_map = kwds.get("vars_to_map", [])
        self._method = kwds.get("method", "direct")

        if self._method == "direct":
            self._mapper = None
        elif self._method == "nearest":
            self._mapper = NearestVal()
        else:
            raise ValueError("method %s not understood" % self._method)

    def initialize(self):
        """Initialize the data mappers."""
        if self._mapper is not None:
            self._mapper.initialize(self._dst, self._src, vars=self._vars_to_map)

    def run(self, stop_time):
        """Map values from one port to another."""
        for (dst_name, src_name) in self._vars_to_map:
            src_values = self._src.get_value(
                src_name, units=self._dst.get_var_units(dst_name)
            )

            if self._mapper is None:
                dst_values = src_values
            else:
                dst_values = self._mapper.run(src_values)

            self._dst.set_value(dst_name, dst_values)

    def finalize(self):
        pass



#! /bin/env python
import numpy as np
import six

from .igrid import CENTERING_CHOICES, CenteringValueError, DimensionError, IField
from .raster import UniformRectilinear
from .rectilinear import Rectilinear
from .structured import Structured
from .unstructured import Unstructured


def combine_args_to_list(*args, **kwds):
    if len(args) == 0:
        args = kwds.get("default", [])
    args = list(args)
    combined_args = []
    for arg in args:
        if isinstance(arg, six.string_types):
            combined_args.append(arg)
        else:
            combined_args.extend(arg)
    return combined_args


class GridField(Unstructured, IField):
    def __init__(self, *args, **kwargs):
        super(GridField, self).__init__(*args, **kwargs)
        self._fields = {}
        self._field_units = {}

    def add_field(
        self,
        field_name,
        val,
        centering="zonal",
        units="-",
        exist_action="clobber",
        time=None,
    ):
        if exist_action not in ("clobber", "append"):
            raise ValueError("'exist_action' must be on of {'clobber', 'append'}")

        try:
            val.shape = val.size
        except AttributeError:
            val = np.array(val)

        if centering not in CENTERING_CHOICES:
            raise CenteringValueError(centering)

        if centering == "zonal" and val.size != self.get_cell_count():
            raise DimensionError(val.size, self.get_cell_count())
        elif centering != "zonal" and val.size != self.get_point_count():
            raise DimensionError(val.size, self.get_point_count())

        self._field_times = {}
        if field_name not in self._fields or exist_action == "clobber":
            self._fields[field_name] = [val]
            self._field_times[field_name] = [time]
        else:
            self._fields[field_name].append(val)
            self._field_times[field_name].append(time)
        self._field_units[field_name] = units

        # self._fields[field_name].shape = val.size

    def pop_field(self, field_name, *args, **kwds):
        return_with_time = kwds.get("return_with_time", False)

        field = self._fields[field_name].pop(*args)
        if return_with_time:
            time = self._field_times[field_name].pop(*args)
            return (time, field)
        else:
            return field

    def get_field(self, field_name):
        if len(self._fields[field_name]) == 1:
            return self._fields[field_name][0]
        else:
            return self._fields[field_name]

    def get_field_units(self, field_name):
        return self._field_units[field_name]

    def set_field_units(self, field_name, units):
        try:
            self._field_units[field_name] = units
        except KeyError:
            pass

    def get_point_fields(self, *args):
        names = combine_args_to_list(*args, default=self._fields.keys())

        fields = {}
        for name in names:
            array = self.get_field(name)
            if array.size == self.get_point_count():
                fields[name] = array
        return fields

        # fields = {}
        # for (field, array) in self._fields.items ():
        #    if array.size == self.get_point_count ():
        #        fields[field] = array
        # return fields

    def get_cell_fields(self, *args):
        names = combine_args_to_list(*args, default=self._fields.keys())

        fields = {}
        for name in names:
            array = self.get_field(name)
            if array.size == self.get_cell_count():
                fields[name] = array
        return fields

    def items(self):
        return self._fields.items()

    def keys(self):
        return self._fields.keys()

    def values(self):
        return self._fields.values()

    def has_field(self, name):
        return name in self._fields


class UnstructuredField(GridField):
    pass


class StructuredField(Structured, GridField):
    def add_field(
        self,
        field_name,
        val,
        centering="zonal",
        units="-",
        exist_action="clobber",
        time=None,
    ):
        if not hasattr(val, "ndim"):
            val = np.array(val)
        # try:
        #    ndim = val.ndim
        # except AttributeError:
        #    val = np.array (val)
        #    ndim = val.ndim

        if centering == "zonal":
            if val.ndim > 1 and np.any(val.shape != self.get_shape() - 1):
                raise DimensionError(val.shape, self.get_shape() - 1)
        elif centering != "zonal":
            if val.ndim > 1 and np.any(val.shape != self.get_shape()):
                raise DimensionError(val.shape, self.get_shape())
        try:
            super(StructuredField, self).add_field(
                field_name,
                val,
                centering=centering,
                units=units,
                exist_action=exist_action,
                time=time,
            )
        except (DimensionError, CenteringValueError):
            raise


class RectilinearField(Rectilinear, StructuredField):
    pass


class RasterField(UniformRectilinear, StructuredField):
    """
    Create a field that looks like this,

    ::

        (0) --- (1) --- (2)
         |       |       |
         |   0   |   1   |
         |       |       |
        (3) --- (4) --- (5)

    Create the field,

        >>> g = RasterField ((2,3), (1,2), (0, 0), indexing='ij')
        >>> g.get_cell_count ()
        2
        >>> g.get_point_count ()
        6

    Add some data at the points of our grid.

        >>> data = np.arange (6)
        >>> g.add_field ('var0', data, centering='point')
        >>> g.get_field ('var0')
        array([0, 1, 2, 3, 4, 5])

    The data can be given either as a 1D array or with the same shape
    as the point grid. In either case, though, it will be flattened.

        >>> data = np.arange (6)
        >>> data.shape = (2, 3)
        >>> g.add_field ('var0', data, centering='point')
        >>> g.get_field ('var0')
        array([0, 1, 2, 3, 4, 5])

    If the size or shape doesn't match, it's an error.

        >>> data = np.arange (2)
        >>> g.add_field ('bad var', data, centering='point') # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
            ...
        DimensionError: 2 != 6

        >>> data = np.ones ((3, 2))
        >>> g.add_field ('bad var', data, centering='point') # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
            ...
        DimensionError: (3, 2) != (2, 3)

    Add data to the cells of the grid. Again, the data can be given
    as a 1D array or with the same shape as the cell grid and if the
    size or shape doesn't match raise an exception.

        >>> data = np.arange (2)
        >>> g.add_field('var1', data, centering='zonal')
        >>> g.get_field('var1')
        array([0, 1])

        >>> data = np.ones((2, 1))
        >>> g.add_field('bad var', data, centering='zonal') # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
            ...
        DimensionError: (2, 1) != (1, 2)

        >>> data = np.arange(3)
        >>> g.add_field('bad var', data, centering='zonal') # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
            ...
        DimensionError: 3 != 2

        >>> data = np.array(3)
        >>> g.add_field('bad var', data, centering='zonal')  # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
            ...
        DimensionError: 1 != 2


    A 1D-Field,

        >>> g = RasterField ((6, ), (1.5, ), (0.5, ), indexing='ij')
        >>> g.get_point_count ()
        6
        >>> g.get_cell_count ()
        5

        >>> point_data = np.arange (6.)
        >>> g.add_field ('Point Data', point_data, centering='point')

        >>> cell_data = np.arange (5.)*10.
        >>> g.add_field ('Cell Data', cell_data, centering='zonal')

        >>> g.get_field ('Cell Data')
        array([  0.,  10.,  20.,  30.,  40.])

        >>> g.get_field ('Point Data')
        array([ 0.,  1.,  2.,  3.,  4.,  5.])


    A 3D-Field,

        >>> g = RasterField ((4, 3, 2), (1.5, 1., 3), (0.5, 0, -.5 ), indexing='ij')
        >>> g.get_point_count ()
        24
        >>> g.get_cell_count ()
        6

        >>> g.add_field ('Point Data', g.get_x (), centering='point')

        >>> cell_data = np.arange (6.)*10.
        >>> g.add_field ('Cell Data', cell_data, centering='zonal')

        >>> g.get_field ('Cell Data')
        array([  0.,  10.,  20.,  30.,  40.,  50.])

        >>> g.get_field('Point Data')
        array([-0.5,  2.5, -0.5,  2.5, -0.5,  2.5, -0.5,  2.5, -0.5,  2.5, -0.5,
                2.5, -0.5,  2.5, -0.5,  2.5, -0.5,  2.5, -0.5,  2.5, -0.5,  2.5,
               -0.5,  2.5])

        >>> g.get_shape()
        array([4, 3, 2])

        >>> g.get_x().size == g.get_field('Point Data').size
        True
        >>> x = g.get_x()
        >>> x.shape
        (24,)
        >>> x.shape = g.get_shape()
        >>> x.shape
        (4, 3, 2)

    """

    pass


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#! /usr/bin/env python
import os
import warnings

from scipy.io import netcdf as nc3

_VALID_NETCDF_FORMATS = set(["NETCDF3_CLASSIC", "NETCDF3_64BIT"])

_NP_TO_NC_TYPE = {
    "float32": "f4",
    "float64": "f8",
    "int8": "i1",
    "int16": "i2",
    "int32": "i4",
    "int64": "i8",
    "uint8": "u1",
    "uint16": "u2",
    "uint32": "u4",
    "uint64": "u8",
}


try:
    import netCDF4 as nc4
except ImportError:
    warnings.warn("Unable to import netCDF4.", ImportWarning)
else:
    _VALID_NETCDF_FORMATS |= set(["NETCDF4_CLASSIC", "NETCDF4"])


def assert_valid_netcdf_format(fmt):
    if fmt not in _VALID_NETCDF_FORMATS:
        raise ValueError(
            "%s: format is one of %s" % (fmt, ", ".join(_VALID_NETCDF_FORMATS))
        )


def open_netcdf(path, mode="r", fmt="NETCDF3_CLASSIC", append=False):
    assert_valid_netcdf_format(fmt)

    if mode != "r":
        if os.path.isfile(path) and append:
            mode = "a"
        else:
            mode = "w"

    if fmt == "NETCDF3_CLASSIC":
        root = nc3.netcdf_file(path, mode, version=1)
    elif fmt == "NETCDF3_64BIT":
        root = nc3.netcdf_file(path, mode, version=2)
    else:
        root = nc4.Dataset(path, mode, format=fmt)

    return root



#! /bin/env python
import numpy as np

from .igrid import IGrid
from .utils import (
    args_as_numpy_arrays,
    coordinates_to_numpy_matrix,
    get_default_coordinate_names,
    get_default_coordinate_units,
)

MAXSIZE = np.iinfo(int).max


class UnstructuredPoints(IGrid):
    def __init__(self, *args, **kwds):
        """
        UnstructuredPoints(x0 [, x1 [, x2]])
        """
        set_connectivity = kwds.pop("set_connectivity", False)
        units = kwds.pop("units", get_default_coordinate_units(len(args)))
        coordinate_names = kwds.pop(
            "coordinate_names", get_default_coordinate_names(len(args))
        )
        self._attrs = kwds.pop("attrs", {})

        if len(args) < 1 or len(args) > 3:
            raise ValueError("number of arguments must be between 1 and 3")

        args = args_as_numpy_arrays(*args)

        self._n_dims = len(args)
        self._point_count = args[0].size

        self._coords = coordinates_to_numpy_matrix(*args)
        self._units = np.array(units)
        self._coordinate_name = np.array(coordinate_names)

        if set_connectivity:
            self._connectivity = np.arange(self._point_count, dtype=int)
            self._offset = self._connectivity + 1
            self._cell_count = 0

        super(UnstructuredPoints, self).__init__()

    def get_attrs(self):
        return self._attrs

    def get_dim_count(self):
        return self._n_dims

    def get_x(self):
        try:
            return self._coords[-1]
        except IndexError:
            raise IndexError("Dimension out of bounds")

    def get_y(self):
        try:
            return self._coords[-2]
        except IndexError:
            raise IndexError("Dimension out of bounds")

    def get_z(self):
        try:
            return self._coords[-3]
        except IndexError:
            raise IndexError("Dimension out of bounds")

    def get_xyz(self):
        return self._coords

    def get_point_coordinates(self, *args, **kwds):
        axis = kwds.pop("axis", None)

        if len(args) == 0:
            inds = np.arange(self.get_point_count())
        elif len(args) == 1:
            inds = args
        else:
            raise ValueError("number of args must be 0 or 1")

        if axis is None:
            axis = np.arange(self.get_dim_count())
            axis.shape = (self.get_dim_count(), 1)

        return tuple(self._coords[axis, inds])

    def get_axis_coordinates(self, axis=None, indexing="ij"):
        if indexing != "ij":
            raise ValueError("only ij indexing is supported")
        if axis is not None and axis >= self.get_dim_count():
            raise ValueError("value for axis exceeds number of dimensions")

        if axis is None:
            return self._coords
        else:
            return self._coords[axis]

    def get_x_units(self):
        return self.get_coordinate_units(-1)

    def get_y_units(self):
        return self.get_coordinate_units(-2)

    def get_z_units(self):
        return self.get_coordinate_units(-3)

    def set_x_units(self, units):
        try:
            self._units[-1] = units
        except IndexError:
            raise IndexError("Dimension out of bounds")

    def set_y_units(self, units):
        try:
            self._units[-2] = units
        except IndexError:
            raise IndexError("Dimension out of bounds")

    def set_z_units(self, units):
        try:
            self._units[-3] = units
        except IndexError:
            raise IndexError("Dimension out of bounds")

    def get_coordinate_units(self, i):
        try:
            return self._units[i]
        except IndexError:
            raise IndexError("Dimension out of bounds")

    def get_coordinate_name(self, i):
        try:
            return self._coordinate_name[i]
        except IndexError:
            raise IndexError("Dimension out of bounds")

    def get_coordinate(self, i):
        try:
            return self._coords[i]
        except IndexError:
            raise IndexError("Dimension out of bounds")

    def get_connectivity(self):
        return self._connectivity

    def get_offset(self):
        return self._offset

    def get_point_count(self):
        return self._point_count

    def get_cell_count(self):
        return self._cell_count

    def get_vertex_count(self):
        return len(self._connectivity)

    def get_max_vertices(self):
        nodes_per_cell = np.diff(self._offset)
        return max(self._offset[0], nodes_per_cell.max())

    def get_connectivity_as_matrix(self, fill_val=MAXSIZE):
        nodes_per_cell = np.diff(self._offset)

        max_vertices = max(self._offset[0], nodes_per_cell.max())
        matrix = np.empty((self._cell_count, max_vertices), dtype=np.int)

        offset = self._offset[0]
        matrix[0, :offset] = self._connectivity[:offset]
        matrix[0, offset:] = fill_val
        for (cell, offset) in enumerate(self._offset[:-1]):
            n_vertices = nodes_per_cell[cell]
            matrix[cell + 1, :n_vertices] = self._connectivity[
                offset : offset + n_vertices
            ]
            matrix[cell + 1, n_vertices:] = fill_val

        return (matrix, fill_val)

    def nodes_per_cell(self):
        offsets = np.empty(self.get_cell_count() + 1, dtype=int)
        (offsets[0], offsets[1:]) = (0, self._offset[:])

        return np.diff(offsets)

    def reverse_element_ordering(self):
        last_offset = 0
        for offset in self._offset:
            c = self._connectivity[last_offset:offset].copy()
            self._connectivity[last_offset:offset] = c[::-1]
            last_offset = offset


class Unstructured(UnstructuredPoints):
    r"""
    Define a grid that consists of two trianges that share two points::

           (2) - (3)
          /   \  /
        (0) - (1)

    Create the grid,

    >>> g = Unstructured([0, 0, 1, 1], [0, 2, 1, 3],
    ...                  connectivity=[0, 2, 1, 2, 3, 1], offset=[3, 6])
    >>> g.get_point_count()
    4
    >>> g.get_cell_count()
    2
    >>> g.get_dim_count()
    2

    >>> g.get_x()
    array([ 0.,  2.,  1.,  3.])

    >>> g.get_y()
    array([ 0.,  0.,  1.,  1.])

    >>> g.get_connectivity()
    array([0, 2, 1, 2, 3, 1])

    >>> g.get_offset()
    array([3, 6])


    Define a grid that consists of points in a line::

        (0) ----- (1) -- (2) - (3)

    Create the grid,

    >>> g = Unstructured ([0., 6., 9., 11.], connectivity=[0, 1, 2, 3], offset=[1, 2, 3, 4])
    >>> g.get_point_count ()
    4
    >>> g.get_cell_count ()
    4

    Eight point that form a unit cube.

    >>> x = [0, 1, 0, 1, 0, 1, 0, 1]
    >>> y = [0, 0, 1, 1, 0, 0, 1, 1]
    >>> z = [0, 0, 0, 0, 1, 1, 1, 1]
    >>> g = Unstructured(z, y, x, connectivity=[0, 1, 2, 3, 4, 5, 6, 7], offset=[8])
    >>> g.get_point_count()
    8
    >>> g.get_cell_count()
    1
    >>> g.get_x()
    array([ 0.,  1.,  0.,  1.,  0.,  1.,  0.,  1.])
    >>> g.get_y()
    array([ 0.,  0.,  1.,  1.,  0.,  0.,  1.,  1.])
    >>> g.get_z()
    array([ 0.,  0.,  0.,  0.,  1.,  1.,  1.,  1.])
    """

    def __init__(self, *args, **kwds):
        set_connectivity = kwds.pop("set_connectivity", True)
        connectivity = kwds.pop("connectivity", None)
        offset = kwds.pop("offset", None)

        if set_connectivity:
            if offset is None:
                offset, args = args[-1], args[:-1]
            if connectivity is None:
                connectivity, args = args[-1], args[:-1]
            self._set_connectivity(connectivity, offset)

        kwds["set_connectivity"] = False
        super(Unstructured, self).__init__(*args, **kwds)

    def _set_connectivity(self, connectivity, offset):
        self._connectivity = np.array(connectivity, dtype=int)
        self._offset = np.array(offset, dtype=int)
        self._connectivity.shape = self._connectivity.size
        self._offset.shape = self._offset.size
        self._cell_count = self._offset.size


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



import contextlib
import os
from functools import partial

import click


out = partial(click.secho, bold=True, err=True)
err = partial(click.secho, fg="red", err=True)


@contextlib.contextmanager
def as_cwd(path):
    prev_cwd = os.getcwd()
    os.chdir(path)
    yield
    os.chdir(prev_cwd)


@contextlib.contextmanager
def suppress_stdout():
    null_fds = [os.open(os.devnull, os.O_RDWR) for x in range(2)]
    # Save the actual stdout (1) and stderr (2) file descriptors.
    save_fds = [os.dup(1), os.dup(2)]

    os.dup2(null_fds[0], 1)
    os.dup2(null_fds[1], 2)

    yield

    # Re-assign the real stdout/stderr back to (1) and (2)
    os.dup2(save_fds[0], 1)
    os.dup2(save_fds[1], 2)
    # Close the null files
    for fd in null_fds + save_fds:
        os.close(fd)



from .celltopoint import CellToPoint
from .mapper import IncompatibleGridError, find_mapper
from .pointtocell import PointToCell
from .pointtopoint import NearestVal

__all__ = [
    "find_mapper",
    "IncompatibleGridError",
    "CellToPoint",
    "NearestVal",
    "PointToCell",
]



#! /bin/env python
import warnings

import yaml

from ...printers.nc.read import field_fromfile
from .interpolate import create_interpolators
from .time_series_names import get_time_series_names


class TimeInterpolator(object):
    def __init__(self):
        self._shape = ()
        self._spacing = ()
        self._origin = ()
        self._input_exchange_items = set()
        self._output_exchange_items = set()
        self._start_time = 0.0
        self._end_time = 0.0
        self._time = 0.0
        self._interpolators = {}

    def initialize(self, source):
        config = read_configuration(source)

        (fields, times) = field_fromfile(config["input_file"], fmt="NETCDF4")

        self._shape = fields[0].get_shape()
        self._spacing = fields[0].get_spacing()
        self._origin = fields[0].get_origin()

        self._input_exchange_items = set()
        self._output_exchange_items = get_time_series_names(fields.keys())

        self._start_time = times[0]
        self._end_time = times[-1]
        self._time = times[0]

        self._interpolators = create_interpolators(
            times, fields, kind=config["interpolation"]
        )

    def update_until(self, time):
        if time < self._start_time or time > self._endtime:
            raise ValueError(
                "time is outside of start/end time ({0} not in [{1}, {2}])".format(
                    time, self._start_time, self._end_time
                )
            )

        self._time = time

    def finalize(self):
        pass

    def get_start_time(self):
        return self._start_time

    def get_end_time(self):
        return self._end_time

    def get_current_time(self):
        return self._time

    def get_input_var_names(self):
        return self._input_exchange_items

    def get_output_var_names(self):
        return self._output_exchange_items

    def get_grid_shape(self, grid_id):
        if grid_id in (None, ""):
            warnings.warn("grid identifier is ignored")
        return self._shape

    def get_grid_spacing(self, grid_id):
        if grid_id in (None, ""):
            warnings.warn("grid identifier is ignored")
        return self._spacing

    def get_grid_origin(self, grid_id):
        if grid_id in (None, ""):
            warnings.warn("grid identifier is ignored")
        return self._origin

    def get_value(self, name):
        return self._interpolators[name](self._time)
        # if name.endswith('_increment'):
        #    name = name[:- len('_increment')]
        #    return (self._interpolators[name](self._time) -
        #            self._interpolators[name](self._last_time))
        # else:
        #    return self._interpolators[name](self._time)


def get_abspath_or_url(filename, prefix=""):
    import os
    from urlparse import urlparse, urlunparse

    parts = urlparse(filename)
    if parts.scheme in ["file", ""]:
        return os.path.join(prefix, parts.path)
    else:
        return urlunparse(parts)


def read_configuration(source):
    config = yaml.safe_load(source)

    input_file = config.get("input_file")
    input_dir = config.get("input_dir", "")
    kind = config.get("interpolation", "linear")

    return {
        "input_file": get_abspath_or_url(input_file, prefix=input_dir),
        "interpolation": kind,
    }


if __name__ == "__main__":
    import doctest

    doctest.testmod()



class Error(Exception):
    """Base class for BMI exceptions"""

    pass


class VarNameError(Error):
    """Exception to indicate a bad input/output variable name"""

    def __init__(self, name):
        self.name = name

    def __str__(self):
        return self.name


class BMI(object):
    def initialize(self, filename):
        pass

    def run(self, time):
        pass

    def finalize(self):
        pass

    def get_input_var_names(self):
        pass

    def get_output_var_names(self):
        pass

    def get_var_grid(self, var_name):
        pass

    def get_var_type(self, var_name):
        pass

    def get_var_units(self, var_name):
        pass

    def get_time_step(self):
        pass

    def get_start_time(self):
        pass

    def get_current_time(self):
        pass

    def get_end_time(self):
        pass

    def get_grid_rank(self, grid_id):
        pass

    def get_grid_spacing(self, grid_id):
        pass

    def get_grid_shape(self, grid_id):
        pass

    def get_grid_x(self, grid_id):
        pass

    def get_grid_y(self, grid_id):
        pass

    def get_grid_z(self, grid_id):
        pass

    def get_grid_connectivity(self, grid_id):
        pass

    def get_grid_offset(self, grid_id):
        pass



"""An empty event.

:class:`PassEvent` implements an event-like interface but doesn't do
anything. This can be a useful event to give to an :class:`EventManager` if
you want to force the manager to stop at particular intervals.
"""


class PassEvent(object):
    """An event that doesn't do anything."""

    def initialize(self):
        pass

    def run(self, stop_time):
        pass

    def finalize(self, stop_time):
        pass



#! /bin/env python
import numpy as np
import yaml


class ConstantScalars(object):
    """Service component that returns scalars."""

    def initialize(self, filename):
        """Initialize the component from a file.

        Parameters
        ----------
        filename : str
            Name of initialization file.
        """
        with open(filename, "r") as opened:
            scalar_vars = yaml.safe_load(opened.read())

        self._vars = {}
        for (name, value) in scalar_vars.items():
            self._vars[name] = np.array(value, dtype=np.float)

        self._shape = (1,)
        self._spacing = (1.0,)
        self._origin = (0.0,)

        self._input_exchange_items = []
        self._output_exchange_items = list(self._vars.keys())

        self._start_time = 0.0
        self._end_time = np.inf
        self._time = 0.0

    def update(self, time):
        """Update one time step."""
        self._time = time

    def update_until(self, time):
        """Update until a time."""
        self._time = time

    def finalize(self):
        """Clean up."""
        pass

    def get_start_time(self):
        """Component start time."""
        return self._start_time

    def get_end_time(self):
        """Component stop time."""
        return self._end_time

    def get_current_time(self):
        """Component current time."""
        return self._time

    def get_input_var_names(self):
        """Input variable names."""
        return self._input_exchange_items

    def get_output_var_names(self):
        """Output variable names."""
        return self._output_exchange_items

    def get_var_grid(self, var):
        """Grid identifier for a variable.

        Parameters
        ----------
        var : str
            Name of grid variable.

        Returns
        -------
        int
            Grid identifier.
        """
        if var in self._vars:
            return 0
        else:
            raise KeyError(var)

    def get_grid_shape(self, grid_id):
        """Shape of grid.

        Parameters
        ----------
        grid_id : int
            Grid identifier.

        Returns
        -------
        shape : tuple
            Shape of the grid.
        """
        if grid_id == 0:
            return self._shape
        else:
            raise KeyError(0)

    def get_grid_spacing(self, grid_id):
        """Spacing of grid.

        Parameters
        ----------
        grid_id : int
            Grid identifier.

        Returns
        -------
        spacing : tuple
            Spacing of nodes in each dimension.
        """
        if grid_id == 0:
            return self._spacing
        else:
            raise KeyError(0)

    def get_grid_origin(self, grid_id):
        """Origin of grid.

        Parameters
        ----------
        grid_id : int
            Grid identifier.

        Returns
        -------
        origin : tuple
            Origin of nodes in each dimension.
        """
        if grid_id == 0:
            return self._origin
        else:
            raise KeyError(0)

    def get_value(self, name):
        """Values on nodes of a grid.

        Parameters
        ----------
        var : str
            Name of grid variable.

        Returns
        -------
        values : ndarray
            Values of the nodes of a grid.
        """
        return self._vars[name]


class River(ConstantScalars):
    pass



#! /bin/env python

import numpy as np
from scipy.spatial import KDTree

from .imapper import IGridMapper, IncompatibleGridError

# from .mapper import IncompatibleGridError


def _flat_view(array):
    return array.view().reshape((array.size,))


def copy_good_values(src, dst, bad_val=-999):
    """Copy only the good values from one array to another.

    Parameters
    ----------
    src : array_like
        Array of source values.
    dst : array_like
        Array into which to copy the good values.
    bad_val : float (optional)
        Value below which indicates a bad value.

    Returns
    -------
    dst : array_like
        The destination array.

    Examples
    --------
    >>> import numpy as np
    >>> from pymt.mappers.pointtopoint import copy_good_values
    >>> x = np.arange(6)
    >>> y = np.zeros((2, 3))
    >>> rv = copy_good_values(x, y)
    >>> rv is y
    True
    >>> y
    array([[ 0.,  1.,  2.],
           [ 3.,  4.,  5.]])
    >>> copy_good_values(x, np.zeros(6), bad_val=3.)
    array([ 0.,  0.,  0.,  0.,  4.,  5.])
    """
    if src.size != dst.size:
        raise ValueError("size mismatch between source and destination arrays")

    flat_src, flat_dst = _flat_view(src), _flat_view(dst)

    ids_to_copy = flat_src > bad_val
    flat_dst[ids_to_copy] = flat_src[ids_to_copy]

    return dst


class NearestVal(IGridMapper):
    """
    Examples
    --------
    >>> import numpy as np
    >>> from pymt.grids.map import RectilinearMap
    >>> from pymt.mappers.pointtopoint import NearestVal
    >>> src = RectilinearMap([0, 1, 2], [0, 2])
    >>> dst = RectilinearMap([.25, 1.25, 2.25], [.25, 1.25])
    >>> mapper = NearestVal()
    >>> mapper.initialize(dst, src)
    >>> mapper.run(np.arange(6.))
    array([ 0.,  1.,  2.,  3.,  4.,  5.])
    """

    _name = "PointToPoint"

    def initialize(self, dest_grid, src_grid, **kwds):
        """Map points on one grid to the nearest points on another.

        Parameters
        ----------
        dest_grid : grid_like
            Grid onto which points are mapped
        src_grid : grid_like
            Grid from which points are taken.
        var_names : iterable of tuples (optional)
            Iterable of (*dest*, *src*) variable names.
        """
        var_names = kwds.get("var_names", None)
        if var_names is None:
            var_names = []

        if len(var_names) > 1:
            raise ValueError("only 0 or 1 var_names allowed")

        if not NearestVal.test(dest_grid, src_grid):
            raise IncompatibleGridError(dest_grid.name, src_grid.name)

        if len(var_names) == 0:
            (x, y) = src_grid.get_x().flat, src_grid.get_y().flat
        else:
            dst_name, src_name = var_names[0]
            (x, y) = (src_grid.get_x(src_name).flat, src_grid.get_y(src_name).flat)

        tree = KDTree(list(zip(x, y)))

        if len(var_names) == 0:
            (x, y) = dest_grid.get_x().flat, dest_grid.get_y().flat
        else:
            dst_name, src_name = var_names[0]
            (x, y) = (dest_grid.get_x(dst_name).flat, dest_grid.get_y(dst_name).flat)

        (_, self._nearest_src_id) = tree.query(list(zip(x, y)))

    def run(self, src_values, **kwds):
        """Map source values onto destination values.

        Parameters
        ----------
        src_values : ndarray
            Source values at points.
        dst_vals : ndarray (optional)
            Destination array to put mapped values.

        Returns
        -------
        dest : ndarray
            The (possibly newly-created) destination array.
        """
        dst_vals = kwds.get("dst_vals", None)
        bad_val = kwds.get("bad_val", -999)

        if dst_vals is None:
            dst_vals = np.zeros(len(self._nearest_src_id), dtype=src_values.dtype)
        elif not isinstance(dst_vals, np.ndarray):
            raise TypeError("Destination array must be a numpy array")

        copy_good_values(
            src_values.flat[self._nearest_src_id], dst_vals, bad_val=bad_val
        )

        return dst_vals

    @staticmethod
    def test(dst_grid, src_grid):
        """Test if grids are compatible with this mapper.

        Parameters
        ----------
        dst_grid : grid_like
            Grid onto which points are mapped
        src_grid : grid_like
            Grid from which points are taken.
        """
        return dst_grid is not None and src_grid is not None

    @property
    def name(self):
        """Name of the grid mapper."""
        return self._name



#! /usr/bin/env python

import os

from .ugrid import close as ugrid_close
from .write import field_tofile


class IDatabase(object):
    def __init__(self):
        pass

    def open(self, path, var_name, **kwds):
        pass

    def write(self, field, **kwds):
        pass

    def close(self):
        pass


def field_changed_size(field, n_points, n_cells):
    return n_points != field.get_point_count() or n_cells != field.get_cell_count()


class Database(IDatabase):
    def open(self, path, var_name, **kwds):
        self.close()

        (root, ext) = os.path.splitext(path)

        self._var_name = var_name
        self._path = path
        self._template = "%s_%%04d%s" % (root, ext)

        self._point_count = None
        self._cell_count = None

    def write(self, field, **kwds):
        kwds.setdefault("append", True)

        if kwds["append"]:
            if self._point_count is not None and self._cell_count is not None:
                if field_changed_size(field, self._point_count, self._cell_count):
                    self._path = self._next_file_name()

            self._point_count = field.get_point_count()
            self._cell_count = field.get_cell_count()
        else:
            self._path = self._next_file_name()

        field_tofile(field, self._path, **kwds)

    def close(self):
        try:
            del self._count
        except AttributeError:
            pass
        try:
            del self._point_count
            del self._cell_count
        except AttributeError:
            pass
        try:
            ugrid_close(self._path)
        except AttributeError:
            pass

    def _next_file_name(self):
        try:
            next_file_name = self._template % self._count
        except AttributeError:
            self._count = 0
            next_file_name = self._template % self._count
        finally:
            self._count += 1

        return next_file_name



#! /bin/env python
"""
Examples
--------

Create a grid that consists of 2x3 nodes using 'xy' and 'ij' indexing.

**ij-indexing**

Create a grid of length 2 in the i direction, and 3 in the j direction.

>>> g = UniformRectilinear ((2,3), (1,2), (.5, 0), indexing='ij', units=('m', 'km'))
>>> g.get_x()
array([ 0.,  2.,  4.,  0.,  2.,  4.])
>>> g.get_y()
array([ 0.5,  0.5,  0.5,  1.5,  1.5,  1.5])
>>> [g.get_x_units(), g.get_y_units()]
['km', 'm']
>>> g.get_z_units()
Traceback (most recent call last):
        ...
IndexError: Dimension out of bounds
>>> [g.get_coordinate_units(i) for i in [0, 1]]
['m', 'km']
>>> g.get_coordinate_units(2)
Traceback (most recent call last):
        ...
IndexError: Dimension out of bounds
>>> g.get_shape()
array([2, 3])
>>> g.get_spacing()
array([ 1.,  2.])
>>> g.get_origin()
array([ 0.5,  0. ])
>>> g.get_offset()
array([4, 8])
>>> g.get_connectivity()
array([0, 1, 4, 3, 1, 2, 5, 4])

**Uniform rectilinear grid of points**

Create a grid of length 2 in the i direction, and 3 in the j direction.

>>> g = UniformRectilinearPoints ((2,3), (1,2), (.5, 0), indexing='ij', set_connectivity=True)
>>> g.get_x()
array([ 0.,  2.,  4.,  0.,  2.,  4.])
>>> g.get_y()
array([ 0.5,  0.5,  0.5,  1.5,  1.5,  1.5])
>>> g.get_shape()
array([2, 3])
>>> g.get_spacing()
array([ 1.,  2.])
>>> g.get_origin()
array([ 0.5,  0. ])

>>> g.get_point_count()
6
>>> g.get_cell_count()
0

The offset runs from 1 up to (and including) the number of points.

>>> all (g.get_offset ()==np.arange (1, g.get_point_count ()+1))
True

The connectivity runs from 0 to one less than the number of points.

>>> all (g.get_connectivity ()==np.arange (g.get_point_count ()))
True


**1D-grid of points**

>>> g = UniformRectilinearPoints ((5, ), (1., ), (.5,), indexing='ij', set_connectivity=True)
>>> g.get_x()
array([ 0.5,  1.5,  2.5,  3.5,  4.5])

**3D-grid of cells**

>>> g = UniformRectilinear ((4, 3, 2), (1, 2, 3), (-1, 0, 1), indexing='ij')
>>> g.get_x()
array([ 1.,  4.,  1.,  4.,  1.,  4.,  1.,  4.,  1.,  4.,  1.,  4.,  1.,
        4.,  1.,  4.,  1.,  4.,  1.,  4.,  1.,  4.,  1.,  4.])
>>> g.get_y()
array([ 0.,  0.,  2.,  2.,  4.,  4.,  0.,  0.,  2.,  2.,  4.,  4.,  0.,
        0.,  2.,  2.,  4.,  4.,  0.,  0.,  2.,  2.,  4.,  4.])
>>> g.get_z()
array([-1., -1., -1., -1., -1., -1.,  0.,  0.,  0.,  0.,  0.,  0.,  1.,
        1.,  1.,  1.,  1.,  1.,  2.,  2.,  2.,  2.,  2.,  2.])
"""

import numpy as np

from .rectilinear import Rectilinear, RectilinearPoints


class UniformRectilinearPoints(RectilinearPoints):
    def __init__(self, shape, spacing, origin, **kwds):
        kwds.setdefault("indexing", "xy")
        kwds.setdefault("set_connectivity", False)

        xi = []
        for (nx, dx, x0) in zip(shape, spacing, origin):
            xi.append(np.arange(nx, dtype=np.float64) * dx + x0)

        self._spacing = np.array(spacing, dtype=np.float64)
        self._origin = np.array(origin, dtype=np.float64)

        super(UniformRectilinearPoints, self).__init__(*xi, **kwds)

    def get_spacing(self):
        """Spacing between nodes in each direction"""
        return self._spacing

    def get_origin(self):
        """Coordinates of the grid's lower-left corner"""
        return self._origin


class UniformRectilinear(UniformRectilinearPoints, Rectilinear):
    """Create a rectilinear grid with uniform spacing in the x and y directions.

    Parameters
    ----------
    shape: tuple of int
        Grid shape measured in number of nodes.
    spacing: tuple of int
        Spacing between nodes in each direction of a rectilinear grid.
    origin: tuple of int
        Coordinates of the lower-left corner of the grid.
    indexing: {'xy', 'ij'}, optional
        Cartesian ('xy', default) or matrix ('ij') indexing of output.
        See Notes for more details.

    Returns
    -------
    UniformRectilinear
        An instance of a UniformRectilinear grid.
    """

    def __init__(self, shape, spacing, origin, **kwds):
        kwds["set_connectivity"] = False
        super(UniformRectilinear, self).__init__(shape, spacing, origin, **kwds)


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



class ChainEvent(object):
    def __init__(self, events):
        self._events = events

    def initialize(self):
        for event in self._events:
            event.initialize()

    def run(self, stop_time):
        for event in self._events:
            event.run(stop_time)

    def update(self, stop_time):
        for event in self._events:
            event.run(stop_time)

    def finalize(self):
        for event in self._events:
            event.finalize()



#! /bin/env python

import numpy as np
from scipy.spatial import KDTree
from six.moves import zip

from .imapper import IGridMapper, IncompatibleGridError

# from .mapper import IncompatibleGridError


def map_points_to_cells(coords, src_grid, src_point_ids, bad_val=-1):
    (dst_x, dst_y) = coords

    point_to_cell_id = np.empty(len(dst_x), dtype=np.int)
    point_to_cell_id.fill(bad_val)

    for (j, point_id) in enumerate(src_point_ids):
        for cell_id in src_grid.get_shared_cells(point_id):
            if src_grid.is_in_cell(dst_x[j], dst_y[j], cell_id):
                point_to_cell_id[j] = cell_id

    return point_to_cell_id


class CellToPoint(IGridMapper):

    _name = "CellToPoint"

    def initialize(self, dest_grid, src_grid, **kwds):
        if not CellToPoint.test(dest_grid, src_grid):
            raise IncompatibleGridError(dest_grid.name, src_grid.name)
        dst_x = dest_grid.get_x()
        dst_y = dest_grid.get_y()

        tree = KDTree(list(zip(src_grid.get_x(), src_grid.get_y())))
        (_, self._nearest_src_id) = tree.query(list(zip(dst_x, dst_y)))

        self._map = map_points_to_cells(
            (dst_x, dst_y), src_grid, self._nearest_src_id, bad_val=-1
        )
        self._bad = self._map == -1

    def run(self, src_values, **kwds):
        dst_vals = kwds.get("dst_vals", None)
        bad_val = kwds.get("bad_val", -999.0)

        if dst_vals is None:
            dst_vals = np.zeros(len(self._map)) + bad_val

        rtn = np.choose(
            src_values.take(self._map) < bad_val, (src_values.take(self._map), dst_vals)
        )
        rtn[self._bad] = bad_val
        dst_vals[:] = rtn

        return rtn

    @staticmethod
    def test(dst_grid, src_grid):
        return all(np.diff(src_grid.get_offset()) > 2)

    @property
    def name(self):
        return self._name



#! /usr/bin/env python
import os

from ...grids import utils as gutils
from .constants import _NP_TO_NC_TYPE, open_netcdf

_OPENED_FILES = {}


def close_all():
    for path in _OPENED_FILES:
        close(path)


def close(path):
    try:
        _OPENED_FILES[path].close()
    except KeyError:
        pass
    else:
        del _OPENED_FILES[path]


class NetcdfField(object):
    def __init__(
        self, path, field, fmt="NETCDF4", append=False, time=None, keep_open=False
    ):
        path = os.path.abspath(path)
        self._path = path
        self._field = field

        if path in _OPENED_FILES and not os.path.isfile(path):
            close(path)

        if path in _OPENED_FILES:
            self._root = _OPENED_FILES[path]
        else:
            self._root = open_netcdf(path, mode="w", fmt=fmt, append=append)

        self._set_mesh_topology()
        self._set_node_variable_data()
        self._set_face_variable_data()
        self._set_time_variable(now=time)

        if keep_open:
            _OPENED_FILES[path] = self._root
        else:
            self.close()

    def _set_mesh_dimensions(self):
        raise NotImplementedError("_set_mesh_dimensions")

    def _set_mesh_coordinate_data(self):
        raise NotImplementedError("_set_mesh_coordinate_data")

    @property
    def node_data_dimensions(self):
        raise NotImplementedError("node_data_dimensions")

    @property
    def face_data_dimensions(self):
        raise NotImplementedError("face_data_dimensions")

    def close(self):
        if self._path in _OPENED_FILES:
            del _OPENED_FILES[self._path]
        self._root.close()

    @property
    def type(self):
        return "unknown"

    @property
    def field(self):
        return self._field

    @property
    def root(self):
        return self._root

    @property
    def topology_dimension(self):
        return len(gutils.non_singleton_shape(self.field))

    @property
    def field_axes(self):
        return gutils.non_singleton_axes(self.field)

    @property
    def node_coordinates(self):
        return ""

    @property
    def face_connectivity(self):
        return "face_nodes_connectivity"

    @property
    def face_node_connectivity(self):
        return "face_nodes"

    @property
    def node_count(self):
        return self.field.get_point_count()

    @property
    def face_count(self):
        return self.field.get_cell_count()

    @property
    def vertex_count(self):
        return self.field.get_vertex_count()

    @property
    def time_count(self):
        try:
            return len(self._root.variables["time"])
        except KeyError:
            return 0

    def has_dimension(self, name):
        return name in self._root.dimensions

    def has_variable(self, name):
        return name in self._root.variables

    def create_dimension(self, name, dim_len):
        try:
            if not self.has_dimension(name):
                self._root.createDimension(name, dim_len)
        except IndexError:
            pass

    def create_variable(self, name, *args, **kwds):
        if not self.has_variable(name):
            self._root.createVariable(name, *args, **kwds)

    def set_variable(self, name, *args, **kwds):
        if len(args) not in (0, 1):
            raise ValueError("number of arguments must be 0 or 1")

        attrs = kwds.pop("attrs", {})

        variable = self.data_variable(name)
        for (attr, value) in attrs.items():
            variable.setncattr(attr, value)

        if len(args) > 0:
            array = args[0]
            if "time" in variable.dimensions:
                n_times = self.time_count
                if array.size > 1:
                    variable[n_times, :] = array.flat
                else:
                    variable[n_times] = array[0]
            else:
                variable[:] = array.reshape(variable.shape)

    def data_variable(self, name):
        return self.root.variables[name]

    def _set_mesh_topology(self):
        self._set_topology()
        self._set_mesh_dimensions()
        self._set_time_dimension()
        self._set_mesh_coordinate_data()
        self._set_face_node_connectivity_data()

    def _set_topology(self):
        self.create_variable("mesh", "i8")
        self.set_variable(
            "mesh",
            attrs={
                "cf_role": "mesh_topology",
                "topology_dimension": self.topology_dimension,
                "node_coordinates": " ".join(self.node_coordinates),
                "face_connectivity": self.face_connectivity,
                "face_node_connectivity": self.face_node_connectivity,
                "type": self.type,
            },
        )

    def _set_time_dimension(self):
        if not self.has_dimension("time"):
            self.create_dimension("time", None)

    def _set_coordinate_data(self):
        self._set_mesh_coordinate_data()
        self._set_face_node_connectivity_data()

    def _set_time_variable(
        self, now=None, units="days", reference="0001-01-01 00:00:00 UTC"
    ):
        self.create_variable("time", "f8", ("time",))

        time = self.data_variable("time")
        time.units = " ".join([units, "since", reference])
        time.long_name = "time"

        if now is not None:
            time[self.time_count - 1] = now
        else:
            time[self.time_count - 1] = self.time_count - 1

    def _set_variable_data(self):
        self._set_node_variable_data()
        self._set_face_variable_data()

    def _set_node_variable_data(self):
        point_fields = self.field.get_point_fields()
        for (var_name, array) in point_fields.items():
            self.create_variable(
                var_name,
                _NP_TO_NC_TYPE[str(array.dtype)],
                ["time"] + list(self.node_data_dimensions),
            )
            self.set_variable(
                var_name,
                array,
                attrs={
                    "units": self.field.get_field_units(var_name),
                    "standard_name": var_name,
                    "long_name": var_name,
                    "location": "node",
                    "coordinates": " ".join(self.node_data_dimensions),
                },
            )

    def _set_face_variable_data(self):
        face_fields = self.field.get_cell_fields()
        for (var_name, array) in face_fields.items():
            self.create_variable(
                var_name,
                _NP_TO_NC_TYPE[str(array.dtype)],
                ["time"] + list(self.face_data_dimensions),
            )
            self.set_variable(
                var_name,
                array,
                attrs={
                    "units": self.field.get_field_units(var_name),
                    "standard_name": var_name,
                    "long_name": var_name,
                    "location": "face",
                    "coordinates": " ".join(self.node_data_dimensions),
                },
            )

    def _set_face_node_connectivity_data(self):
        pass


class NetcdfRectilinearField(NetcdfField):
    @property
    def type(self):
        return "rectilinear"

    @property
    def node_coordinates(self):
        return gutils.non_singleton_dimension_names(self.field)

    @property
    def topology_dimension(self):
        return len(gutils.non_singleton_shape(self.field))

    @property
    def node_data_dimensions(self):
        return gutils.non_singleton_dimension_names(self.field)

    @property
    def axis_coordinates(self):
        return gutils.non_singleton_dimension_names(self.field)

    def _set_mesh_dimensions(self):
        field_shape = self.field.get_shape()
        for (name, axis) in zip(self.axis_coordinates, self.field_axes):
            self.create_dimension(name, field_shape[axis])

    def _set_mesh_coordinate_data(self):
        for (name, axis) in zip(self.axis_coordinates, self.field_axes):
            self.create_variable(name, "f8", (name,))
            self.set_variable(
                name,
                self.field.get_axis_coordinates(axis=axis),
                attrs={
                    "units": self.field.get_coordinate_units(axis),
                    "standard_name": self.field.get_coordinate_name(axis),
                    "long_name": self.field.get_coordinate_name(axis),
                    "name": self.field.get_coordinate_name(axis),
                },
            )


class NetcdfStructuredField(NetcdfRectilinearField):
    @property
    def type(self):
        return "structured"

    @property
    def node_data_dimensions(self):
        return gutils.non_singleton_dimension_names(self._field)

    @property
    def node_coordinates(self):
        return gutils.non_singleton_dimension_names(self._field)
        # names = gutils.non_singleton_dimension_names(self._field)
        # return ['node_' + name for name in names]

    def _set_mesh_dimensions(self):
        NetcdfRectilinearField._set_mesh_dimensions(self)

        for name in self.node_coordinates:
            self.create_dimension(name, self.node_count)

    def _set_mesh_coordinate_data(self):
        dims = self.node_data_dimensions
        # for (name, axis) in zip(self.node_coordinates, self.field_axes):
        for (name, axis) in zip(self.node_data_dimensions, self.field_axes):
            self.create_variable(name, "f8", dims)
            self.set_variable(
                name,
                self.field.get_coordinate(axis),
                attrs={
                    "units": self.field.get_coordinate_units(axis),
                    "standard_name": self.field.get_coordinate_name(axis),
                    "long_name": self.field.get_coordinate_name(axis),
                },
            )


class NetcdfUnstructuredField(NetcdfStructuredField):
    @property
    def type(self):
        return "unstructured"

    @property
    def node_coordinates(self):
        names = []
        for axis in range(self.field.get_dim_count()):
            names.append("node_" + self.field.get_coordinate_name(axis))
        return names

    @property
    def topology_dimension(self):
        return self.field.get_dim_count()

    @property
    def node_data_dimensions(self):
        return ("n_node",)
        # dimensions = []
        # for name in self.node_coordinates:
        #    dimensions.append((name, ))
        # return dimensions

    @property
    def face_data_dimensions(self):
        return ("n_face",)

    def _set_mesh_dimensions(self):
        # for name in self.node_coordinates:
        #    self.create_dimension(name, self.node_count)

        self.create_dimension("n_node", self.node_count)
        self.create_dimension("n_face", self.face_count)
        self.create_dimension("n_vertex", self.vertex_count)
        self.create_dimension("n_max_face_nodes", self.field.get_max_vertices())

    def _set_mesh_coordinate_data(self):
        dims = self.node_data_dimensions
        # for (name, axis) in zip(self.node_data_dimensions, self.field_axes):
        # for (name, axis) in zip(self.node_coordinates, self.field_axes):
        for (axis, name) in enumerate(self.node_coordinates):
            self.create_variable(name, "f8", dims)
            self.set_variable(
                name,
                self.field.get_coordinate(axis),
                attrs={
                    "units": self.field.get_coordinate_units(axis),
                    "standard_name": self.field.get_coordinate_name(axis),
                    "long_name": self.field.get_coordinate_name(axis),
                },
            )

    def _set_face_node_connectivity_data(self):
        self.create_variable("face_nodes_connectivity", "i8", ("n_vertex",))
        self.set_variable(
            "face_nodes_connectivity",
            self.field.get_connectivity(),
            attrs={
                "cf_role": "face_node_connectivity",
                "long_name": "Maps every face to its corner nodes.",
                "start_index": 0,
            },
        )

        self.create_variable("face_nodes_offset", "i8", ("n_face",))
        self.set_variable(
            "face_nodes_offset",
            self.field.get_offset(),
            attrs={
                "cf_role": "face_node_offset",
                "long_name": "Maps face index into connectivity array",
            },
        )

        (connectivity, fill_val) = self.field.get_connectivity_as_matrix()

        self.create_variable(
            "face_nodes", "i8", ("n_face", "n_max_face_nodes"), fill_value=fill_val
        )
        self.set_variable(
            "face_nodes",
            connectivity,
            attrs={
                "cf_role": "face_node_connectivity",
                "long_name": "Maps every face to its corner nodes.",
                "start_index": 0,
            },
        )



#! /usr/bin/env python

import six
from six.moves.configparser import ConfigParser

from ..framework import services
from ..printers.nc.database import Database as NcDatabase

# from ..printers.vtk.vtu import Database as VtkDatabase
from ..utils.prefix import names_with_prefix, strip_prefix
from .utils import (
    construct_file_name,
    construct_port_as_field,
    next_unique_file_name,
    reconstruct_port_as_field,
)


class PortPrinter(object):
    _format = ""
    _printer_class = None

    def __init__(self, port, var_name, filename=None):
        if isinstance(port, six.string_types):
            self._port = services.get_component_instance(port)
        else:
            self._port = port
        self._var_name = var_name
        self._filename = filename
        if filename is None:
            self._filename = var_name

        self._field = construct_port_as_field(self._port, var_name)
        self._printer = self._printer_class()  # pylint: disable=not-callable

    @property
    def var_name(self):
        return self._var_name

    @property
    def format(self):
        return self._format

    def open(self, clobber=False):
        file_name = construct_file_name(self._filename, fmt=self.format, prefix="")
        if not clobber:
            file_name = next_unique_file_name(file_name)

        self._printer.open(file_name, self.var_name)

    def close(self):
        self._printer.close()
        self._printer = self._printer_class()  # pylint: disable=not-callable

    def write(self):
        self.resync_field_to_port()
        self._printer.write(self._field)

    def resync_field_to_port(self):
        self._field = reconstruct_port_as_field(self._port, self._field)

    @classmethod
    def from_string(cls, source, prefix="print"):
        config = ConfigParser()
        config.readfp(six.StringIO(source))
        return cls._from_config(config, prefix=prefix)

    @classmethod
    def from_path(cls, path, prefix="print"):
        config = ConfigParser()
        config.read(path)
        return cls._from_config(config, prefix=prefix)

    @classmethod
    def _from_config(cls, config, prefix="print"):
        printers = []
        for section in names_with_prefix(config.sections(), prefix):
            printers.append(
                cls.from_dict(
                    {
                        "port": config.get(section, "port"),
                        "format": config.get(section, "format"),
                        "name": strip_prefix(section, prefix),
                    }
                )
            )
        if len(printers) == 1:
            return printers[0]
        else:
            return printers

    @classmethod
    def from_dict(cls, d):
        try:
            printer_class = _FORMAT_TO_PRINTER[d["format"]]
        except KeyError:
            raise ValueError("%s: unknown printer format" % d["format"])
        return printer_class(
            d["port"], d["name"], filename=d.get("filename", d["name"])
        )


# class VtkPortPrinter(PortPrinter):
#     _format = "vtk"
#     _printer_class = VtkDatabase


class NcPortPrinter(PortPrinter):
    _format = "nc"
    _printer_class = NcDatabase


_FORMAT_TO_PRINTER = {
    # "vtk": VtkPortPrinter,
    "nc": NcPortPrinter,
    "netcdf": NcPortPrinter,
}



#! /usr/bin/env python
import matplotlib.pyplot as plt
import numpy as np


def quick_plot(bmi, name, **kwds):
    if bmi.var_location(name) == "none":
        raise ValueError(f"{name} does not have an associated grid to plot")

    gid = bmi.var_grid(name)
    gtype = bmi.grid_type(gid)
    grid = bmi.grid[gid]

    x_label = "{name} ({units})".format(
        name=grid.node_x.standard_name, units=grid.node_x.units
    )
    y_label = "{name} ({units})".format(
        name=grid.node_y.standard_name, units=grid.node_y.units
    )

    z = bmi.get_value(name)

    if gtype.startswith("unstructured"):
        x, y = grid.node_x.values, grid.node_y.values
        nodes_per_face = bmi.grid_nodes_per_face(gid)
        if np.all(nodes_per_face == 3):
            tris = bmi.grid_face_nodes(gid).reshape((-1, 3))
            # tris = bmi.grid_face_node_connectivity(gid).reshape((-1, 3))
            plt.tripcolor(x, y, tris, z, **kwds)
        else:
            raise ValueError(
                "quickplot is only able to plot unstructured meshes of triangles"
            )
    elif gtype in ("uniform_rectilinear", "structured_quad"):
        shape = bmi.grid_shape(gid)
        spacing = bmi.grid_spacing(gid)
        origin = bmi.grid_origin(gid)
        x = np.arange(shape[-1]) * spacing[-1] + origin[-1]
        y = np.arange(shape[-2]) * spacing[-2] + origin[-2]
        plt.pcolormesh(x, y, z.reshape(shape), **kwds)
    else:
        raise ValueError("no plotter for {gtype}".format(gtype=gtype))

    plt.axis("tight")
    plt.gca().set_aspect("equal")
    plt.xlabel(x_label)
    plt.ylabel(y_label)

    cbar = plt.colorbar()
    cbar.ax.set_ylabel("{name} ({units})".format(name=name, units=bmi.var_units(name)))



#! /bin/env python

import numpy as np

from ..grids.esmp import EsmpUnstructuredField
from .imapper import IGridMapper
from .mapper import IncompatibleGridError

try:
    import ESMF as esmf
except ImportError:
    esmf = None
    REGRID_METHOD = None
    UNMAPPED_ACTION = None
else:
    REGRID_METHOD = esmf.RegridMethod.CONSERVE
    UNMAPPED_ACTION = esmf.UnmappedAction.ERROR


class EsmpMapper(IGridMapper):

    _name = None

    @property
    def name(self):
        return self._name

    @staticmethod
    def test(dst_grid, src_grid):
        raise NotImplementedError("test")

    def init_fields(self):
        raise NotImplementedError("init_fields")

    def initialize(self, dest_grid, src_grid, **kwds):
        method = kwds.get("method", REGRID_METHOD)
        unmapped = kwds.get("unmapped", UNMAPPED_ACTION)

        if not EsmpMapper.test(dest_grid, src_grid):
            raise IncompatibleGridError(dest_grid.name, src_grid.name)

        self._src = EsmpUnstructuredField(
            src_grid.get_x(),
            src_grid.get_y(),
            src_grid.get_connectivity(),
            src_grid.get_offset(),
        )
        self._dst = EsmpUnstructuredField(
            dest_grid.get_x(),
            dest_grid.get_y(),
            dest_grid.get_connectivity(),
            dest_grid.get_offset(),
        )

        self.init_fields()
        self._regridder = esmf.Regrid(
            self.get_source_field(),
            self.get_dest_field(),
            regrid_method=method,
            unmapped_action=unmapped,
        )

    def run(self, src_values, **kwds):
        dest_values = kwds.get("dest_values", None)
        src_ptr = self.get_source_data()
        src_ptr[:] = src_values
        # src_ptr.data = src_values

        # dst_ptr = ESMP.ESMP_FieldGetPtr(dst_field)
        dst_ptr = self.get_dest_data()

        if dest_values is not None:
            dst_ptr[:] = dest_values
            # dst_ptr.data = dest_values
        else:
            dst_ptr.fill(0.0)

        # ESMP.ESMP_FieldRegrid(self.get_source_field(), self.get_dest_field(),
        #                      self._routehandle)
        self._regridder(self.get_source_field(), self.get_dest_field())

        return dest_values

    def finalize(self):
        pass
        # ESMP.ESMP_FieldRegridRelease(self._routehandle)

    def get_source_field(self):
        return self._src.as_esmp("src")

    def get_dest_field(self):
        return self._dst.as_esmp("dst")

    def get_source_data(self):
        return self._src.as_esmp("src").data
        # return ESMP.ESMP_FieldGetPtr(self.get_source_field())

    def get_dest_data(self):
        return self._dst.as_esmp("dst").data
        # return ESMP.ESMP_FieldGetPtr(self.get_dest_field())


class EsmpCellToCell(EsmpMapper):

    _name = "CellToCell"

    def init_fields(self):
        data = np.empty(self._src.get_cell_count(), dtype=np.float64)
        self._src.add_field("src", data, centering="zonal")

        data = np.empty(self._dst.get_cell_count(), dtype=np.float64)
        self._dst.add_field("dst", data, centering="zonal")

    @staticmethod
    def test(dst_grid, src_grid):
        return all(np.diff(dst_grid.get_offset()) > 2) and all(
            np.diff(src_grid.get_offset()) > 2
        )


class EsmpPointToPoint(EsmpMapper):

    _name = "PointToPoint"

    def init_fields(self):
        data = np.empty(self._src.get_point_count(), dtype=np.float64)
        self._src.add_field("src", data, centering="point")

        data = np.empty(self._dst.get_point_count(), dtype=np.float64)
        self._dst.add_field("dst", data, centering="point")

    @staticmethod
    def test(dst_grid, src_grid):
        return dst_grid is not None and src_grid is not None



from collections import OrderedDict


def prefix_is_empty(prefix):
    """Check if a namespace prefix is empty.

    A prefix is empty if it is ``None``, just a "." or an empty string.

    Return ``True`` if empty, otherwise ``False``.
    """
    return prefix is None or prefix == "." or len(prefix) == 0


def names_with_prefix(names, prefix):
    """Find names that begin with a common *prefix*. In this case, names
    are a series ``.``-separated words, much like module names.

    Returns a ``list`` of all names that begin with *prefix*. The order of
    matched names is maintained.

    >>> names_with_prefix(['foo.bar', 'foobar.baz'], 'foo')
    ['foo.bar']
    >>> names_with_prefix(['foo.bar', 'foo.bar', 'foo.foo'], 'foo')
    ['foo.bar', 'foo.foo']
    """
    if prefix_is_empty(prefix):
        return set(names)

    if not prefix.endswith("."):
        prefix = prefix + "."

    matching_names = OrderedDict()
    for name in names:
        if name.startswith(prefix):
            matching_names[name] = None

    return list(matching_names.keys())


def strip_prefix(name, prefix):
    """Remove a prefix from a name, including any leading "."s.

    >>> strip_prefix('foo.bar', 'foo')
    'bar'
    >>> strip_prefix('foo.bar', '')
    'foo.bar'
    """
    if prefix_is_empty(prefix):
        return name

    if not prefix.endswith("."):
        prefix += "."

    if name.startswith(prefix):
        return name[len(prefix) :]
    else:
        raise ValueError("%s does not start with %s" % (name, prefix))



#! /usr/bin/env python
from warnings import warn

from ...grids.assertions import is_rectilinear, is_structured
from .ugrid import (
    NetcdfRectilinearField,
    NetcdfStructuredField,
    NetcdfUnstructuredField,
)


def field_tofile(
    field,
    path,
    append=False,
    attrs=None,
    time=None,
    time_units=None,
    time_reference=None,
    long_name=None,
    fmt="NETCDF4",
    keep_open=False,
):
    if time_units is not None:
        warn('ignoring keyword "time_units"', UserWarning)
    if time_reference is not None:
        warn('ignoring keyword "time_reference"', UserWarning)
    if not keep_open:
        warn('resetting "keep_open" to True', UserWarning)

    attrs = attrs or {}
    long_name = long_name or {}
    args = (path, field)
    kwds = dict(append=append, fmt=fmt, time=time, keep_open=True)

    if is_rectilinear(field, strict=False):
        NetcdfRectilinearField(*args, **kwds)
    elif is_structured(field, strict=False):
        NetcdfStructuredField(*args, **kwds)
    else:
        NetcdfUnstructuredField(*args, **kwds)



from ..portprinter.port_printer import PortPrinter
from ..utils import as_cwd


class PrintEvent(object):
    def __init__(self, *args, **kwds):
        # self._printer = PortPrinter.from_dict(kwds)
        self._run_dir = kwds.pop("run_dir", ".")
        self._kwds = kwds

    def initialize(self, *args):
        with as_cwd(self._run_dir):
            self._printer = PortPrinter.from_dict(self._kwds)
            self._printer.open()

    def run(self, time):
        with as_cwd(self._run_dir):
            self._printer.write()

    def finalize(self):
        with as_cwd(self._run_dir):
            self._printer.close()



class GridType(object):
    _type = None

    def __eq__(self, that):
        return isinstance(that, self.__class__) or str(self) == str(that)

    def __str__(self):
        return self._type

    def __repr__(self):
        return "%s()" % self.__class__.__name__


class GridTypeRectilinear(GridType):
    _type = "rectilinear"


class GridTypeStructured(GridType):
    _type = "structured"


class GridTypeUnstructured(GridType):
    _type = "unstructured"



"""
Notes
-----
This function supports both indexing conventions through the indexing keyword
argument.  Giving the string 'ij' returns a meshgrid with matrix indexing,
while 'xy' returns a meshgrid with Cartesian indexing.  In the 2-D case
with inputs of length M and N, the outputs are of shape (N, M) for 'xy'
indexing and (M, N) for 'ij' indexing.  In the 3-D case with inputs of
length M, N and P, outputs are of shape (N, M, P) for 'xy' indexing and (M,
N, P) for 'ij' indexing.  The difference is illustrated by the following
code snippet::

        xv, yv = meshgrid(x, y, sparse=False, indexing='ij')
        for i in range(nx):
            for j in range(ny):
                # treat xv[i,j], yv[i,j]

        xv, yv = meshgrid(x, y, sparse=False, indexing='xy')
        for i in range(nx):
            for j in range(ny):
                # treat xv[j,i], yv[j,i]

See Also
--------
index_tricks.mgrid : Construct a multi-dimensional "meshgrid"
                     using indexing notation.
index_tricks.ogrid : Construct an open multi-dimensional "meshgrid"
                     using indexing notation.

Examples
--------

>>> nx, ny = (3, 2)
>>> x = np.linspace(0, 1, nx)
>>> y = np.linspace(0, 1, ny)
>>> xv, yv = meshgrid(x, y)
>>> xv
array([[ 0. ,  0.5,  1. ],
       [ 0. ,  0.5,  1. ]])
>>> yv
array([[ 0.,  0.,  0.],
       [ 1.,  1.,  1.]])
>>> xv, yv = meshgrid(x, y, sparse=True)  # make sparse output arrays
>>> xv
array([[ 0. ,  0.5,  1. ]])
>>> yv
array([[ 0.],
       [ 1.]])

`meshgrid` is very useful to evaluate functions on a grid.

>>> x = np.arange(-5, 5, 0.1)
>>> y = np.arange(-5, 5, 0.1)
>>> xx, yy = meshgrid(x, y, sparse=True)
>>> z = np.sin(xx**2 + yy**2) / (xx**2 + yy**2)
"""

import numpy as np


# Based on scitools meshgrid
def meshgrid(*xi, **kwargs):
    """Coordinate matrices from two or more coordinate vectors.

    Make N-D coordinate arrays for vectorized evaluations of
    N-D scalar/vector fields over N-D grids, given
    one-dimensional coordinate arrays x1, x2,..., xn.

    Parameters
    ----------
    xi: array_like
            1-D arrays representing the coordinates of a grid.
    indexing: {'xy', 'ij'}, optional
        Cartesian ('xy', default) or matrix ('ij') indexing of output.
        See Notes for more details.
    sparse: bool, optional
        If True a sparse grid is returned in order to conserve memory.
        Default is False.
    copy: bool, optional
        If False, a view into the original arrays are returned in
        order to conserve memory.  Default is True.  Please note that
        ``sparse=False, copy=False`` will likely return non-contiguous arrays.
        Furthermore, more than one element of a broadcast array may refer to
        a single memory location.  If you need to write to the arrays, make
        copies first.

    Returns
    -------
    ndarray
        For vectors `x1`, `x2`,..., 'xn' with lengths ``Ni=len(xi)`` ,
        return ``(N1, N2, N3,...Nn)`` shaped arrays if indexing='ij'
        or ``(N2, N1, N3,...Nn)`` shaped arrays if indexing='xy'
        with the elements of `xi` repeated to fill the matrix along
        the first dimension for `x1`, the second for `x2` and so on.

    """
    if len(xi) < 2:
        msg = "meshgrid() takes 2 or more arguments (%d given)" % int(len(xi) > 0)
        raise ValueError(msg)

    args = np.atleast_1d(*xi)
    ndim = len(args)

    copy_ = kwargs.get("copy", True)
    sparse = kwargs.get("sparse", False)
    indexing = kwargs.get("indexing", "xy")
    if indexing not in ["xy", "ij"]:
        raise ValueError("Valid values for `indexing` are 'xy' and 'ij'.")

    s0 = (1,) * ndim
    output = [x.reshape(s0[:i] + (-1,) + s0[i + 1 : :]) for i, x in enumerate(args)]

    shape = [x.size for x in output]

    if indexing == "xy":
        # switch first and second axis
        output[0].shape = (1, -1) + (1,) * (ndim - 2)
        output[1].shape = (-1, 1) + (1,) * (ndim - 2)
        shape[0], shape[1] = shape[1], shape[0]

    if sparse:
        if copy_:
            return [x.copy() for x in output]
        else:
            return output
    else:
        # Return the full N-D matrix (not only the 1-D vector)
        if copy_:
            mult_fact = np.ones(shape, dtype=int)
            return [x * mult_fact for x in output]
        else:
            return np.broadcast_arrays(*output)


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



import importlib
import os
import socket
import sys

import six
import yaml
from model_metadata.find import find_model_data_files


class redirect(object):
    def __init__(self, stdout=None, stderr=None):
        self._stdout = sys.stdout
        self._stderr = sys.stderr

        self._out = stdout or sys.stdout
        self._err = stderr or sys.stderr

    def __enter__(self):
        if isinstance(self._out, six.string_types):
            sys.stdout = open(self._out, "w")
        else:
            sys.stdout = self._out
        if isinstance(self._err, six.string_types):
            sys.stderr = open(self._err, "w")
        else:
            sys.stderr = self._err

    def __exit__(self, type_, value, traceback):
        if isinstance(self._out, six.string_types):
            sys.stdout.close()
        if isinstance(self._err, six.string_types):
            sys.stderr.close()

        sys.stdout = self._stdout
        sys.stderr = self._stderr


def read_component_configuration(names, vars=None):  # pylint: disable=redefined-builtin
    module = importlib.import_module(".".join(["pymt", "models"]))

    names = names or module.__all__
    vars_ = vars or ("uses", "provides", "info", "parameters", "files", "api")

    configs = {}

    for name in names:
        model = module.__dict__[name]()

        config = {}
        if "uses" in vars_:
            config["uses"] = list(model.get_input_var_names())
        if "provides" in vars_:
            config["provides"] = list(model.get_output_var_names())
        if "info" in vars_:
            with open(os.path.join(model.datadir, "info.yaml")) as fp:
                config["info"] = yaml.safe_load(fp)
        if "parameters" in vars_:
            with open(os.path.join(model.datadir, "parameters.yaml")) as fp:
                config["parameters"] = yaml.safe_load(fp)
        if "files" in vars_:
            config["files"] = find_model_data_files(model.datadir)
        if "api" in vars_:
            with open(os.path.join(model.datadir, "api.yaml")) as fp:
                api = yaml.safe_load(fp)
            config["api"] = {
                "module": module.__name__,
                "class": name,
                "initialize_args": api.get("initialize_args", ""),
            }
        configs[name] = config

    return configs


def main():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("name", nargs="*", help="Name of model")
    parser.add_argument("--vars", default=None, help="List of variables to get")
    parser.add_argument("--output", type=argparse.FileType("w"), help="Output file")

    args = parser.parse_args()

    vars_ = args.vars
    if vars_ is not None:
        vars_ = args.vars.split(",")

    with redirect(stdout=sys.stderr):
        config = read_component_configuration(args.name, vars=vars_)

    host = {
        "hostname": socket.gethostname(),
        "os_name": os.uname()[0],
        "os_release": os.uname()[2],
        "os_version": os.uname()[3],
        "platform": os.uname()[4],
        "prefix": sys.exec_prefix,
        "environ": {
            "PATH": os.pathsep.join(
                [os.path.join(sys.exec_prefix, "bin"), "/usr/bin", "/bin"]
            )
        },
        "python_version": ".".join(
            [
                str(sys.version_info.major),
                str(sys.version_info.minor),
                str(sys.version_info.micro),
            ]
        ),
    }

    print(
        yaml.dump({"host": host, "components": config}, default_flow_style=False),
        file=args.output,
    )


if __name__ == "__main__":
    main()



from collections import OrderedDict

import numpy as np
import xarray as xr

from landlab.graph import StructuredQuadGraph, UniformRectilinearGraph, RectilinearGraph

COORDINATE_NAMES = ["z", "y", "x"]
INDEX_NAMES = ["k", "j", "i"]


def index_names(rank):
    return ["node_" + ind for ind in INDEX_NAMES[-rank:]]


def coordinate_names(rank):
    return ["node_" + d for d in COORDINATE_NAMES[-rank:]]


class _Base(xr.Dataset):
    __slots__ = "bmi", "grid_id", "grid_type", "ndim", "metadata"

    def __init__(self, bmi, grid_id):
        self.bmi = bmi
        self.grid_id = grid_id
        self.grid_type = bmi.grid_type(grid_id)
        self.ndim = bmi.grid_ndim(grid_id)
        self.metadata = OrderedDict()
        super(_Base, self).__init__()

    def set_mesh(self):
        self.update({"mesh": xr.DataArray(data=self.grid_id, attrs=self.metadata)})

    def set_shape(self, shape):
        self.update({"node_shape": xr.DataArray(data=shape, dims=("rank",))})

    def set_spacing(self, spacing):
        self.update({"node_spacing": xr.DataArray(data=spacing, dims=("rank",))})

    def set_origin(self, origin):
        self.update({"node_origin": xr.DataArray(data=origin, dims=("rank",))})

    def get_nodes(self):
        nodes = []
        for dim_name in COORDINATE_NAMES[: -(self.ndim + 1) : -1]:
            data = getattr(self.bmi, "grid_" + dim_name)(self.grid_id)
            nodes.insert(0, data)

        return tuple(nodes)

    def set_nodes(self):
        coords = {}
        for dim_name in COORDINATE_NAMES[: -(self.ndim + 1) : -1]:
            data = getattr(self.bmi, "grid_" + dim_name)(self.grid_id)
            coord = xr.DataArray(
                data=data,
                dims=("node",),
                attrs={"standard_name": dim_name, "units": "m"},
            )
            coords["node_" + dim_name] = coord

        self.update(coords)

    def set_nodes_rectilinear(self, grid_coords):
        coords_at_node = np.meshgrid(*grid_coords, indexing="ij")
        coords = {}
        for axis, dim_name in enumerate(COORDINATE_NAMES[-self.ndim :]):
            coord = xr.DataArray(
                data=coords_at_node[axis].reshape(-1),
                dims=("node",),
                attrs={"standard_name": dim_name, "units": "m"},
            )
            coords["node_" + dim_name] = coord

        self.update(coords)

    def set_connectivity(self, data=None):
        if data is None:
            data = self.bmi.grid_face_nodes(self.grid_id)
        face_node_connectivity = xr.DataArray(
            data=data,
            dims=("vertex",),
            attrs={"standard_name": "Face-node connectivity"},
        )
        self.update({"face_node_connectivity": face_node_connectivity})

    def set_offset(self, data=None):
        if data is None:
            data = self.bmi.grid_face_node_offset(self.grid_id)
        face_node_offset = xr.DataArray(
            data=data,
            dims=("face",),
            attrs={"standard_name": "Offset to face-node connectivity"},
        )
        self.update({"face_node_offset": face_node_offset})


class Scalar(_Base):
    __slots__ = ()

    def __init__(self, *args):
        super(Scalar, self).__init__(*args)

        if self.ndim != 0:
            raise ValueError("scalar must be rank 0")

        self.metadata = OrderedDict(
            [
                ("cf_role", "grid_topology"),
                ("long_name", "scalar value"),
                ("topology_dimension", self.ndim),
                ("type", self.grid_type),
            ]
        )
        self.set_mesh()


class Vector(_Base):
    __slots__ = ()

    def __init__(self, *args):
        super(Vector, self).__init__(*args)

        if self.ndim != 1:
            raise ValueError("vector must be rank 1")

        self.metadata = OrderedDict(
            [
                ("cf_role", "grid_topology"),
                ("long_name", "vector value"),
                ("topology_dimension", self.ndim),
                ("type", self.grid_type),
            ]
        )
        self.set_mesh()


class Points(_Base):
    __slots__ = ()

    def __init__(self, *args):
        super(Points, self).__init__(*args)

        self.metadata = OrderedDict(
            [
                ("cf_role", "mesh_topology"),
                ("long_name", "Topology data of 2D unstructured points"),
                ("topology_dimension", self.ndim),
                ("node_coordinates", "node_x node_y"),
                ("type", self.grid_type),
            ]
        )
        self.set_mesh()
        self.set_nodes()


class Unstructured(_Base):
    __slots__ = ()

    def __init__(self, *args):
        super(Unstructured, self).__init__(*args)

        self.metadata = OrderedDict(
            [
                ("cf_role", "mesh_topology"),
                ("long_name", "Topology data of 2D unstructured points"),
                ("topology_dimension", self.ndim),
                ("node_coordinates", "node_x node_y"),
                ("type", self.grid_type),
            ]
        )
        self.set_mesh()
        self.set_nodes()
        self.set_connectivity()
        self.set_offset()


class StructuredQuadrilateral(_Base):
    __slots__ = ()

    def __init__(self, *args):
        super(StructuredQuadrilateral, self).__init__(*args)

        if self.ndim < 1 or self.ndim > 3:
            raise ValueError("structured_quadrilateral grid must be rank 1, 2, or 3")

        shape = self.bmi.grid_shape(self.grid_id)

        self.metadata = OrderedDict(
            [
                ("cf_role", "grid_topology"),
                (
                    "long_name",
                    "Topology data of {}D structured quadrilateral".format(self.ndim),
                ),
                ("topology_dimension", self.ndim),
                ("node_coordinates", " ".join(coordinate_names(self.ndim))),
                ("node_dimensions", " ".join(index_names(self.ndim))),
                ("type", self.grid_type),
            ]
        )
        self.set_mesh()
        self.set_shape(shape)
        nodes = self.get_nodes()
        self.set_nodes()

        graph = StructuredQuadGraph(nodes, shape=tuple(shape))

        self.set_connectivity(data=graph.nodes_at_patch.reshape((-1,)))
        self.set_offset(
            data=np.arange(1, graph.number_of_patches + 1, dtype=np.int32) * 4
        )


class Rectilinear(_Base):
    __slots__ = ()

    def __init__(self, *args):
        super(Rectilinear, self).__init__(*args)

        if self.ndim < 1 or self.ndim > 3:
            raise ValueError("rectilinear grid must be rank 1, 2, or 3")

        shape = self.bmi.grid_shape(self.grid_id)

        self.metadata = OrderedDict(
            [
                ("cf_role", "grid_topology"),
                (
                    "long_name",
                    "Topology data of {}D structured quadrilateral".format(self.ndim),
                ),
                ("topology_dimension", self.ndim),
                ("node_coordinates", " ".join(coordinate_names(self.ndim))),
                ("node_dimensions", " ".join(index_names(self.ndim))),
                ("type", self.grid_type),
            ]
        )
        self.set_mesh()
        self.set_shape(shape)
        nodes = self.get_nodes()
        self.set_nodes_rectilinear(nodes)

        graph = RectilinearGraph(nodes)

        self.set_connectivity(data=graph.nodes_at_patch.reshape((-1,)))
        self.set_offset(
            data=np.arange(1, graph.number_of_patches + 1, dtype=np.int32) * 4
        )


class UniformRectilinear(_Base):
    __slots__ = ()

    def __init__(self, *args):
        super(UniformRectilinear, self).__init__(*args)

        if self.ndim < 1 or self.ndim > 3:
            raise ValueError("uniform_rectangular grid must be rank 1, 2, or 3")

        shape = self.bmi.grid_shape(self.grid_id)
        spacing = self.bmi.grid_spacing(self.grid_id)
        origin = self.bmi.grid_origin(self.grid_id)

        self.metadata = OrderedDict(
            [
                ("cf_role", "grid_topology"),
                (
                    "long_name",
                    "Topology data of {}D structured quadrilateral".format(self.ndim),
                ),
                ("topology_dimension", self.ndim),
                ("node_coordinates", " ".join(coordinate_names(self.ndim))),
                ("node_dimensions", " ".join(index_names(self.ndim))),
                ("node_spacing", "node_spacing"),
                ("node_origin", "node_origin"),
                ("type", self.grid_type),
            ]
        )
        self.set_mesh()
        self.set_shape(shape)
        self.set_spacing(spacing)
        self.set_origin(origin)

        grid_coords = []
        for dim in range(self.ndim):
            grid_coords.append(
                np.arange(shape[dim], dtype=float) * spacing[dim] + origin[dim]
            )
        self.set_nodes_rectilinear(grid_coords)

        graph = UniformRectilinearGraph(shape, spacing=spacing, origin=origin)

        self.set_connectivity(data=graph.nodes_at_patch.reshape((-1,)))
        self.set_offset(
            data=np.arange(1, graph.number_of_patches + 1, dtype=np.int32) * 4
        )


def dataset_from_bmi_grid(bmi, grid_id):
    grid_type = bmi.grid_type(grid_id)
    if grid_type == "points":
        grid = Points(bmi, grid_id)
    elif grid_type == "uniform_rectilinear":
        grid = UniformRectilinear(bmi, grid_id)
    elif grid_type == "structured_quadrilateral":
        grid = StructuredQuadrilateral(bmi, grid_id)
    elif grid_type == "scalar":
        grid = Scalar(bmi, grid_id)
    elif grid_type.startswith("unstructured"):
        grid = Unstructured(bmi, grid_id)
    elif grid_type == "vector":
        grid = Vector(bmi, grid_id)
    elif grid_type == "rectilinear":
        grid = Rectilinear(bmi, grid_id)
    else:
        raise ValueError("grid type not understood ({gtype})".format(gtype=grid_type))

    return grid



from collections import defaultdict

from scipy import interpolate

from .time_series_names import sort_time_series_names


def get_time_series_values(field, ordering="ascending", prefix=""):
    values = defaultdict(list)

    names = sort_time_series_names(field, ordering=ordering, prefix=prefix)

    for (var_name, ordered_names) in names.items():
        for name in ordered_names:
            values[var_name].append(field.get_values(name))

    return values


def get_field_values(fields, names):
    values = []
    for name in names:
        values.append(fields.get_values(name))
    return values


def create_interpolators(times, fields, prefix="", kind="linear"):
    interpolators = {}

    names = sort_time_series_names(fields.keys(), prefix=prefix, ordering="ascending")

    for (var_name, ordered_names) in names.items():
        interpolators[var_name] = create_interpolator(
            fields, zip(ordered_names, times), kind=kind
        )

    return interpolators


def create_interpolator(fields, time_stamps, kind="linear"):
    names, times = zip(*time_stamps)

    values = get_field_values(fields, names)
    return interpolate.interp1d(times, values, axis=0, kind=kind)



__all__ = []

import sys

from .model_collection import ModelCollection

for name, cls in ModelCollection().items():
    __all__.append(name)
    setattr(sys.modules[__name__], name, cls)

try:
    del name, cls
except NameError:
    pass
del sys, ModelCollection



#! /bin/env python
"""
Examples
========

Point-to-point Mapping
----------------------

>>> import numpy as np
>>> from pymt.grids.map import RectilinearMap as Rectilinear

>>> src = Rectilinear([0, 1, 2], [0, 2])
>>> dst = Rectilinear([.5, 1.5, 2.5], [.25, 1.25])

>>> src.get_x()
array([ 0.,  2.,  0.,  2.,  0.,  2.])

>>> src.get_y()
array([ 0.,  0.,  1.,  1.,  2.,  2.])

>>> dst.get_x()
array([ 0.25,  1.25,  0.25,  1.25,  0.25,  1.25])
>>> dst.get_y()
array([ 0.5,  0.5,  1.5,  1.5,  2.5,  2.5])

>>> src_vals = np.arange(src.get_point_count(), dtype=np.float64)

Map the source values on the source points to the destination grid
using nearest neighbor.

>>> from pymt.mappers import NearestVal

>>> mapper = NearestVal()
>>> mapper.initialize(dst, src)
>>> mapper.run(src_vals)
array([ 0.,  1.,  2.,  3.,  4.,  5.])

>>> mappers = find_mapper(dst, src)
>>> len(mappers)
3
>>> mappers[0].name
'PointToPoint'

>>> src_vals[2] = -999
>>> dst_vals = np.full(dst.get_point_count(), -1.)
>>> mapper.run(src_vals, dst_vals=dst_vals)
array([ 0.,  1., -1.,  3.,  4.,  5.])

Cell-to-point Mapping
---------------------

The source grid looks like,

::

    (0) ------ (1)
     |          |
     |          |
    (2) ------ (3)
     |          |
     |          |
    (4) ------ (5)
     |          |
     |          |
    (7) ------ (7)

>>> from pymt.mappers import CellToPoint
>>> from pymt.grids.map import UniformRectilinearMap as UniformRectilinear
>>> from pymt.grids.map import UnstructuredPointsMap as UnstructuredPoints

>>> (dst_x, dst_y) = (np.array([.45, 1.25, 3.5]), np.array([.75, 2.25, 3.25]))

>>> src = UniformRectilinear((2,4), (2, 1), (0, 0))
>>> dst = UnstructuredPoints(dst_x, dst_y)

>>> src_vals = np.arange(src.get_cell_count(), dtype=np.float64)

>>> mapper = CellToPoint()
>>> mapper.initialize(dst, src)
>>> mapper.run(src_vals, bad_val=-999)
array([   0.,    2., -999.])

>>> src_vals = np.arange(src.get_cell_count(), dtype=np.float64)
>>> src_vals[0] = -9999
>>> dst_vals = np.zeros(dst.get_point_count()) + 100
>>> _ = mapper.run(src_vals, dst_vals=dst_vals)
>>> dst_vals
array([ 100.,    2., -999.])

Point-to-cell Mapping
---------------------

>>> from pymt.mappers.pointtocell import PointToCell
>>> (src_x, src_y) = (np.array ([.45, 1.25, 3.5, .0, 1.]),
...                   np.array ([.75, 2.25, 3.25, .9, 1.1]))

>>> src = UnstructuredPoints(src_x, src_y)
>>> dst = UniformRectilinear((2,4), (2, 1), (0, 0))

>>> src_vals = np.arange(src.get_point_count(), dtype=np.float64)

>>> mapper = PointToCell()
>>> mapper.initialize(dst, src)

>>> mapper.run(src_vals, bad_val=-999)
array([ 1.5,  4. ,  1. ])

>>> mapper.run(src_vals, bad_val=-999, method=np.sum)
array([ 3.,  4.,  1.])

>>> src_vals[0] = -9999
>>> dst_vals = np.zeros(dst.get_cell_count ()) - 1
>>> _ = mapper.run(src_vals, dst_vals=dst_vals)
>>> dst_vals
array([-1.,  4.,  1.])


Point on cell edges
-------------------

>>> (src_x, src_y) = (np.array ([0, .5, 1., 2, 3.5]),
...                   np.array ([1., 1., .0, 3, 3.]))
>>> src = UnstructuredPoints(src_x, src_y)
>>> dst = UniformRectilinear((2,4), (2, 1), (0, 0))

>>> mapper = PointToCell()
>>> mapper.initialize(dst, src)

>>> src_vals = np.arange(src.get_point_count(), dtype=np.float64)
>>> dst_vals = np.zeros(dst.get_cell_count()) - 1
>>> _ = mapper.run(src_vals, dst_vals=dst_vals)
>>> dst_vals
array([ 1. ,  0.5,  3. ])

A big mapper
============

>>> (m, n) = (20, 40)
>>> (src_x, src_y) = np.meshgrid(range(m), range(n))
>>> src = UnstructuredPoints(src_y, src_x)
>>> dst = UniformRectilinear((n + 1, m + 1), (1, 1), (-.5, -.5))

>>> mapper = PointToCell()
>>> mapper.initialize(dst, src)

>>> src_vals = np.arange(src.get_point_count(), dtype=np.float64)
>>> dst_vals = np.zeros(dst.get_cell_count(), dtype=np.float64) - 1
>>> _ = mapper.run(src_vals, dst_vals=dst_vals)

>>> from numpy.testing import assert_array_equal
>>> assert_array_equal(dst_vals, src_vals)
"""

from .celltopoint import CellToPoint
from .imapper import IncompatibleGridError
from .pointtocell import PointToCell
from .pointtopoint import NearestVal

_MAPPERS = [NearestVal, CellToPoint, PointToCell]


def find_mapper(dst_grid, src_grid):
    """Find appropriate mappers to map bewteen two grid-like objects"""
    choices = []
    for cls in _MAPPERS:
        if cls.test(dst_grid, src_grid):
            choices.append(cls())

    if len(choices) == 0:
        raise IncompatibleGridError(dst_grid.name, src_grid.name)

    return choices



import numpy as np


def transform_math_to_azimuth(angle, units="rad"):
    angle *= -1.0
    if units == "rad":
        angle += np.pi * 0.5
    else:
        angle += 90.0
    return angle


def transform_azimuth_to_math(angle, units="rad"):
    angle *= -1.0
    if units == "rad":
        angle += np.pi * 0.5
    else:
        angle += 90.0
    return angle



import pkg_resources

from gimli import UnitSystem
from .model_collection import ModelCollection

MODELS = ModelCollection()

__version__ = pkg_resources.get_distribution("pymt").version
__all__ = ["UnitSystem", "MODELS"]

del pkg_resources



import numpy as np


class UniformRectilinearGridPort(object):
    def __init__(self):
        self._shape = (4, 5)
        self._spacing = (1.0, 2.0)
        self._origin = (0.0, 1.0)
        self._values = {
            "landscape_surface__elevation": np.ones(self._shape),
            "sea_surface__temperature": np.zeros(self._shape),
            "sea_floor_surface_sediment__mean_of_grain_size": np.zeros(self._shape),
            "air__density": np.zeros(self._shape),
            "glacier_top_surface__slope": np.zeros(self._shape),
        }

    def get_var_grid(self, var_name):
        if var_name in self._values:
            return 0
        else:
            raise KeyError(var_name)

    def get_grid_shape(self, grid_id):
        if grid_id == 0:
            return self._shape
        else:
            raise KeyError(grid_id)

    def get_grid_spacing(self, grid_id):
        if grid_id == 0:
            return self._spacing
        else:
            raise KeyError(grid_id)

    def get_grid_origin(self, grid_id):
        if grid_id == 0:
            return self._origin
        else:
            raise KeyError(grid_id)

    def get_value(self, var_name):
        return self._values[var_name]



"""Manage the execution of events on a timeline.

The :class:`EventManager` puts events on a :class:`~pymt.timeline.Timeline` and
then
orchestrates the setup, execution, and teardown of the events as the time line
advances. Events are classes that implement an event interface. That is, they
implement the following methods:

* `initialize()`
* `run(time)`
* `finalize()`

Because the :class:`EventManager` class itself is event-like, it is able to
manage other :class:`EventManager` instances and even handle recursive events.

Examples
--------
Create an event that prints, "hello".

>>> class PrintHello(object):
...     def initialize(self):
...         print("hello from initialize")
...     def run(self, time):
...         print("hello!")
...     def finalize(self):
...         print("hello from finalize")

Add an instance of the event to be run at a regular interval to the manager.

>>> mngr = EventManager([(PrintHello(), 2.)])
>>> mngr.initialize()
hello from initialize
>>> mngr.run(1.)
>>> mngr.run(2.)
hello!
>>> mngr.finalize()
hello from finalize

You can also use an :class:`EventManager` as a context,

>>> with EventManager([(PrintHello(), 2.)]) as mngr:
...     mngr.run(4.)
hello from initialize
hello!
hello!
hello from finalize
"""
from six import StringIO
from six.moves.configparser import ConfigParser

from ..timeline import Timeline
from ..utils.prefix import names_with_prefix


class EventManager(object):
    """
    Parameters
    ----------
    events : dict-like
        Events as event-object/repeat interval pairs.

    See Also
    --------
    :class:`pymt.timeline.Timeline`

    Examples
    --------
    Create an :class:`EventManager` without any events.

    >>> mngr = EventManager()
    >>> mngr.time
    0.0
    >>> len(mngr)
    0

    Create a manager with one recurring event.

    >>> from pymt.events.empty import PassEvent
    >>> mngr = EventManager([(PassEvent(), 1.)])
    >>> len(mngr)
    1

    Create a manager with two recurring events.

    >>> mngr = EventManager([(PassEvent(), 1.), (PassEvent(), 2.)])
    >>> len(mngr)
    2
    """

    def __init__(self, *args):
        if len(args) > 1:
            raise TypeError(
                "__init__() takes 1 or 2 arguments (%d given)" % (len(args) + 1,)
            )

        self._timeline = Timeline(*args)
        self._initializing = False
        self._initialized = False
        self._running = False
        self._finalizing = False

        self._order = list(*args)

    def initialize(self):
        """Initialize the managed events.

        Execute the `initialize` methods of each of the events that are being
        managed, making sure the manager is not initialized if it is already
        in the initialization process. Events are initialized in the order
        they were given at creation.
        """
        if not self._initializing:
            self._initializing = True
            for (event, _) in self._order:
                try:
                    event.initialize()
                except Exception:
                    print("error initializing")
                    print(event)
                    raise
            self._initialized = True

    def run(self, stop_time):
        """Run events until some time.

        Execute the `run` method of each of the events being managed until
        *stop_time* is reached on the time line.

        Parameters
        ----------
        stop_time : float
            Time to run events until.
        """
        self.initialize()
        if not self._running:
            self._running = True
            for event in self._timeline.iter_until(stop_time):
                try:
                    event.run
                except AttributeError:
                    event.update(self._timeline.time)
                else:
                    event.run(self._timeline.time)
            self._running = False

    def finalize(self):
        """Finalize managed events.

        Execute the `finalize` method for each of the events being managed.
        Events are finalized in the reverse order in which they were
        initialized, and only if the manager has not yet been initialized.
        """
        if self._initialized:
            if not self._finalizing:
                self._finalizing = True
                # for event in self._timeline.events:
                for (event, _) in self._order[::-1]:
                    event.finalize()
            self._initialized = False

    def add_recurring_event(self, event, interval):
        """Add a managed event.

        Add *event* to the list of managed events and run it with the
        recurrence interval, *interval*.

        Parameters
        ----------
        event : event-like
            Event to be managed.
        interval : float
            Recurrence interval for the event.
        """
        self._timeline.add_recurring_event(event, interval)
        self._order.append((event, interval))

    @property
    def time(self):
        """Current time along the time line."""
        return self._timeline.time

    @classmethod
    def from_string(cls, source, prefix=""):
        """Create an `EventManager` from a string.

        Parameters
        ----------
        source : str
            Ini-formatted string.
        prefix : str
            Prefix for section.

        Returns
        -------
        EventManager
            A newly-created `EventManager`.

        See Also
        --------
        :meth:`from_path` : Alternate constructor that uses a path name.
        """
        config = ConfigParser()
        config.readfp(StringIO(source))
        return cls._from_config(config, prefix=prefix)

    @classmethod
    def from_path(cls, path, prefix=""):
        """Create an `EventManager` from a file.

        Parameters
        ----------
        path : str
            Path to ini-formatted file.
        prefix : str
            Prefix for section.

        Returns
        -------
        EventManager
            A newly-created `EventManager`.

        See Also
        --------
        :meth:`from_string` : Alternate constructor that uses a string.
        """
        config = ConfigParser()
        config.read(path)
        return cls._from_config(config, prefix=prefix)

    @classmethod
    def _from_config(cls, config, prefix=""):
        event_names = names_with_prefix(config.sections(), prefix)
        events = []
        for name in event_names:
            events.append((name, config.get(name, "interval")))
        return cls(events)

    def __enter__(self):
        self.initialize()
        return self

    def __exit__(self, exception_type, value, traceback):
        self.finalize()

    def __len__(self):
        return len(self._order)



import numpy as np

from pymt.grids import UniformRectilinearPoints


def get_instance(name):
    return _INSTANCES[name]


def instantiate(cls, name):
    if name in _INSTANCES:
        raise ValueError(name)
    else:
        _INSTANCES[name] = get_class(cls)()


def get_class(name):
    return _CLASSES[name]


def get_class_names():
    return _CLASSES.keys()


# class EmptyPort(object):
class EmptyPort(UniformRectilinearPoints):
    _name = None

    def __init__(self):
        UniformRectilinearPoints.__init__(self, (4, 5), (1.0, 2.0), (0.0, 1.0))
        # self._shape = (4, 5)
        # self._spacing = (1., 2.)
        # self._origin = (0., 1.)
        self._values = {}
        self._time = self.start_time

    def get_component_name(self):
        return self._name

    def initialize(self):
        for array in self._values.values():
            array.fill(0.0)

    def run(self, time):
        self._time = time
        for array in self._values.values():
            array.fill(self.current_time)

    def finalize(self):
        for array in self._values.values():
            array.fill(0.0)

    def get_var_grid(self, var_name):
        if var_name in self._values:
            return 0
        else:
            raise KeyError(var_name)

    def get_var_units(self, var_name):
        if var_name in self._values:
            return "-"
        else:
            raise KeyError(var_name)

    def get_grid_shape(self, grid_id):
        if grid_id == 0:
            return self.get_shape()
        else:
            raise KeyError(grid_id)

    def get_grid_spacing(self, grid_id):
        if grid_id == 0:
            return self.get_spacing()
        else:
            raise KeyError(grid_id)

    def get_grid_origin(self, grid_id):
        if grid_id == 0:
            return self.get_origin()
        else:
            raise KeyError(grid_id)

    def get_value(self, var_name, units=None):
        try:
            return self._values[var_name]
        except KeyError:
            print(self._values.keys())
            raise

    def set_value(self, var_name, values):
        try:
            self._values[var_name].flat = values.flat
        except KeyError:
            print(self._values.keys())
            raise

    @property
    def start_time(self):
        return 0.0

    @property
    def current_time(self):
        return self._time

    @property
    def end_time(self):
        return 100.0

    @property
    def time_step(self):
        return 1.0


class WaterPort(EmptyPort):
    _name = "water_port"

    def __init__(self):
        EmptyPort.__init__(self)
        self._values = {
            "ocean_surface__temperature": np.empty(self._shape),
            "ocean_surface__density": np.empty(self._shape),
        }


class AirPort(EmptyPort):
    _name = "air_port"

    def __init__(self):
        EmptyPort.__init__(self)
        self._values = {
            "air__temperature": np.empty(self._shape),
            "air__density": np.empty(self._shape),
        }


class EarthPort(EmptyPort):
    _name = "earth_port"

    def __init__(self):
        EmptyPort.__init__(self)
        self._values = {
            "earth_surface__temperature": np.empty(self._shape),
            "earth_surface__density": np.empty(self._shape),
            "glacier_top_surface__slope": np.empty(self._shape),
        }


_SERVICES = {
    "air_port": AirPort(),
    "water_port": WaterPort(),
    "earth_port": EarthPort(),
}

_CLASSES = {"air_port": AirPort, "water_port": WaterPort, "earth_port": EarthPort}
_INSTANCES = {}



#! /bin/env python


class MapperError(Exception):
    """Base class for error in this package."""

    pass


class IncompatibleGridError(MapperError):
    """Error to indicate that the source grid cannot be mapped to the
    destination.
    """

    def __init__(self, dst, src):
        self._src = src
        self._dst = dst

    def __str__(self):
        return "Unable to map %s to %s" % (self._src, self._dst)


class NoMapperError(MapperError):
    def __init__(self, dst, src):
        self._src = src
        self._dst = dst

    def __str__(self):
        return "No mapper to map %s to %s" % (self._src, self._dst)


class IGridMapper(object):
    """Interface for a grid mapper."""

    def initialize(self, dest_grid, src_grid, **kwds):
        """Initialize the mapper to map from a source grid to a destination
        grid.
        """
        pass

    def run(self, src_values, **kwds):
        """Map values on the source grid to the destination grid."""
        pass



from .assertions import (
    is_rectilinear,
    is_structured,
    is_uniform_rectilinear,
    is_unstructured,
)
from .field import RasterField, RectilinearField, StructuredField, UnstructuredField
from .igrid import DimensionError
from .raster import UniformRectilinear, UniformRectilinearPoints
from .rectilinear import Rectilinear, RectilinearPoints
from .structured import Structured, StructuredPoints
from .unstructured import Unstructured, UnstructuredPoints

__all__ = [
    "UniformRectilinear",
    "UniformRectilinearPoints",
    "Rectilinear",
    "RectilinearPoints",
    "Structured",
    "StructuredPoints",
    "Unstructured",
    "UnstructuredPoints",
    "RasterField",
    "RectilinearField",
    "StructuredField",
    "UnstructuredField",
    "is_uniform_rectilinear",
    "is_rectilinear",
    "is_structured",
    "is_unstructured",
    "DimensionError",
]



#! /bin/env python

"""
Examples
--------

Create a grid of length 2 in the x direction, and 3 in the y direction.

>>> g = Rectilinear([1., 2., 3.], [1., 2., 4., 8.])
>>> g.get_point_count()
12
>>> g.get_cell_count()
6
>>> g.get_x()
array([ 1.,  2.,  4.,  8.,  1.,  2.,  4.,  8.,  1.,  2.,  4.,  8.])
>>> g.get_y()
array([ 1.,  1.,  1.,  1.,  2.,  2.,  2.,  2.,  3.,  3.,  3.,  3.])
>>> g.get_shape()
array([3, 4])

Create a grid of length 2 in the i direction, and 3 in the j direction.

>>> g = Rectilinear([1., 2., 4., 8.], [1., 2., 3.], indexing='ij')
>>> g.get_x()
array([ 1.,  2.,  3.,  1.,  2.,  3.,  1.,  2.,  3.,  1.,  2.,  3.])
>>> g.get_y()
array([ 1.,  1.,  1.,  2.,  2.,  2.,  4.,  4.,  4.,  8.,  8.,  8.])
>>> g.get_shape()
array([4, 3])

>>> g.get_offset()
array([ 4,  8, 12, 16, 20, 24])
>>> g.get_connectivity()
array([ 0,  1,  4,  3,  1,  2,  5,  4,  3,  4,  7,  6,  4,  5,  8,  7,  6,
        7, 10,  9,  7,  8, 11, 10])

**Rectilinear grid of points**

Create a grid of length 2 in the i direction, and 3 in the j direction.

>>> g = RectilinearPoints ([1., 2., 4., 8.], [1., 2., 3.], indexing='ij', set_connectivity=True)
>>> g.get_x()
array([ 1.,  2.,  3.,  1.,  2.,  3.,  1.,  2.,  3.,  1.,  2.,  3.])
>>> g.get_y()
array([ 1.,  1.,  1.,  2.,  2.,  2.,  4.,  4.,  4.,  8.,  8.,  8.])
>>> g.get_shape ()
array([4, 3])
>>> g.get_point_count()
12
>>> g.get_cell_count()
0

The offset runs from 1 up to (and including) the number of points.

>>> all (g.get_offset ()==np.arange (1, g.get_point_count ()+1))
True

The connectivity runs from 0 to one less than the number of points.

>>> all (g.get_connectivity ()==np.arange (g.get_point_count ()))
True


**1D Rectilinear grid**

>>> g = Rectilinear([1,3,4,5,6], set_connectivity=True)
>>> g.get_x()
array([ 1.,  3.,  4.,  5.,  6.])
>>> g.get_point_count()
5
>>> g.get_cell_count()
4
>>> g.get_connectivity()
array([0, 1, 1, 2, 2, 3, 3, 4])
>>> g.get_offset()
array([2, 4, 6, 8])


**3D Rectilinear grid**

>>> g = Rectilinear ([0, 1], [2, 3], set_connectivity=True, indexing='ij')
>>> g.get_x()
array([ 2.,  3.,  2.,  3.])
>>> g.get_y()
array([ 0.,  0.,  1.,  1.])

>>> g = Rectilinear ([0, 1], [2, 3], [4, 5], set_connectivity=True, indexing='ij')
>>> g.get_x()
array([ 4.,  5.,  4.,  5.,  4.,  5.,  4.,  5.])
>>> g.get_y()
array([ 2.,  2.,  3.,  3.,  2.,  2.,  3.,  3.])
>>> g.get_z()
array([ 0.,  0.,  0.,  0.,  1.,  1.,  1.,  1.])
>>> g.get_point_count()
8
>>> g.get_cell_count()
1

>>> g = Rectilinear([0, 1, 2, 3], [4, 5, 6], [7, 8], set_connectivity=True, indexing='ij')
>>> g.get_x()
array([ 7.,  8.,  7.,  8.,  7.,  8.,  7.,  8.,  7.,  8.,  7.,  8.,  7.,
        8.,  7.,  8.,  7.,  8.,  7.,  8.,  7.,  8.,  7.,  8.])
>>> g.get_shape()
array([4, 3, 2])
>>> x = g.get_x()
>>> x.shape = g.get_shape()
"""

import numpy as np

from pymt.grids.meshgrid import meshgrid

from .structured import Structured, StructuredPoints


class RectilinearPoints(StructuredPoints):
    def __init__(self, *args, **kwds):
        kwds.setdefault("set_connectivity", False)
        kwds.setdefault("indexing", "xy")

        n_dim = len(args)
        if n_dim < 1 or n_dim > 3:
            raise ValueError("number of dimensions must be between 1 and 3")

        coords = []
        for arg in args:
            coords.append(np.array(arg, dtype=np.float64))

        if n_dim > 1:
            XI = meshgrid(*coords, indexing="ij")
        else:
            XI = [np.array(args[0], dtype=np.float64)]

        shape = XI[0].shape

        self._x_coordinates = np.array(coords[-1], dtype=np.float64)
        try:
            self._y_coordinates = np.array(coords[-2], dtype=np.float64)
        except IndexError:
            pass

        try:
            self._z_coordinates = np.array(coords[-3], dtype=np.float64)
        except IndexError:
            pass

        args = XI + [shape]
        super(RectilinearPoints, self).__init__(*args, **kwds)

    def get_x_coordinates(self):
        """

        Examples
        --------

        >>> g = Rectilinear([0, 1], [2, 3], [4, 5], set_connectivity=True, indexing='ij')
        >>> g.get_x_coordinates()
        array([ 4.,  5.])
        """
        return self._x_coordinates

    def get_y_coordinates(self):
        """
        Examples
        --------

        >>> g = Rectilinear([0, 1], [2, 3], [4, 5], set_connectivity=True, indexing='ij')
        >>> g.get_y_coordinates()
        array([ 2.,  3.])
        """
        return self._y_coordinates

    def get_z_coordinates(self):
        """
        Examples
        --------

        >>> g = Rectilinear([0, 1], [2, 3], [4, 5], set_connectivity=True, indexing='ij')
        >>> g.get_z_coordinates()
        array([ 0.,  1.])
        """
        return self._z_coordinates

    def get_xyz_coordinates(self):
        return (
            self.get_x_coordinates(),
            self.get_y_coordinates(),
            self.get_z_coordinates(),
        )

    def get_axis_coordinates(self, axis=None, indexing="ij"):
        if indexing != "ij":
            raise ValueError("only ij indexing is supported")

        if axis is None:
            return [
                self.get_z_coordinates(),
                self.get_y_coordinates(),
                self.get_x_coordinates(),
            ]
        else:
            n_dims = self.get_dim_count()
            if axis >= n_dims:
                raise ValueError("axis is greater than number of dimensions")

            axis = n_dims - axis - 1

            if axis == 0:
                return self.get_x_coordinates()
            elif axis == 1:
                return self.get_y_coordinates()
            elif axis == 2:
                return self.get_z_coordinates()


class Rectilinear(RectilinearPoints, Structured):
    """Create a rectilinear grid.

    Parameters
    ----------
    x: ndarray
        1-D array of x-coordinates of nodes.
    y: ndarray
        1-D array of y-coordinates of nodes.
    shape: tuple of int
        Grid shape measured in number of nodes
    indexing: {'xy', 'ij'}, optional
        Cartesian ('xy', default) or matrix ('ij') indexing of output.

    Returns
    -------
    Rectilinear
        An instance of a Rectilinear grid.
    """

    def __init__(self, *args, **kwds):
        kwds["set_connectivity"] = True
        super(Rectilinear, self).__init__(*args, **kwds)


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#! /usr/bin/env


def is_callable_method(obj, method):
    return hasattr(obj, method) and callable(getattr(obj, method))


def assert_is_uniform_rectilinear(grid):
    if not is_uniform_rectilinear(grid):
        raise AssertionError("Grid is not uniform rectilinear")


def assert_is_rectilinear(grid, strict=True):
    if not is_rectilinear(grid, strict=strict):
        if strict:
            raise AssertionError("Grid is not strictly rectilinear")
        else:
            raise AssertionError("Grid is not rectilinear")


def assert_is_structured(grid, strict=True):
    if not is_structured(grid, strict=strict):
        if strict:
            raise AssertionError("Grid is not strictly structured")
        else:
            raise AssertionError("Grid is not structured")


def assert_is_unstructured(grid, strict=True):
    if not is_unstructured(grid, strict=True):
        if strict:
            raise AssertionError("Grid is not strictly unstructured")
        else:
            raise AssertionError("Grid is not unstructured")


def is_uniform_rectilinear(grid):
    return (
        is_callable_method(grid, "get_spacing")
        and is_callable_method(grid, "get_origin")
        and is_callable_method(grid, "get_shape")
    )


def is_rectilinear(grid, strict=True):
    loosely_rectilinear = is_callable_method(grid, "get_shape") and is_callable_method(
        grid, "get_xyz_coordinates"
    )
    if not strict:
        return loosely_rectilinear
    else:
        return loosely_rectilinear and not is_uniform_rectilinear(grid)


def is_structured(grid, strict=True):
    loosely_structured = is_callable_method(grid, "get_shape")

    if not strict:
        return loosely_structured
    else:
        return loosely_structured and not (
            is_rectilinear(grid) or is_uniform_rectilinear(grid)
        )


def is_unstructured(grid, strict=True):
    if not strict:
        return True
    else:
        return not (
            is_structured(grid) or is_rectilinear(grid) or is_uniform_rectilinear(grid)
        )



#! /usr/bin/env python
import bisect

from scipy.interpolate import interp1d

_MINIMUM_SIZE_FOR_METHOD = {
    "linear": 2,
    "nearest": 2,
    "zero": 2,
    "slinear": 2,
    "quadratic": 3,
    "cubic": 4,
    "previous": 2,
    "next": 2,
}


class TimeInterpolator(object):

    METHODS = (
        "linear",
        "nearest",
        "zero",
        "slinear",
        "quadratic",
        "cubic",
        "previous",
        "next",
    )

    def __init__(
        self, data=(), method="linear", fill_value="extrapolate", maxsize=None
    ):
        """Interpolate data based on an evolving time series of data values.

        Parameters
        ---------
        data : iterable of (*time*, *data*), optional
            The data to use for the interpolation as an iterable of (*time*, *data*).
            *data* can either be scalars or numpy arrays.
        method : str, optional
            The interpolation method to use.
        fill_value : str or float, optional
            The value to use if trying to interpolate values that are outside the
            stored times. The default is to extrapolate to the given time.
        maxsize : int, optional
            The maximum size of the buffer that holds the interpolation times. When
            the buffer reaches this size, the oldest times will be popped off a the
            stack.
        """
        self._data = []
        self._time = []
        self._func = None
        self._method = None
        self._fill_value = None
        self._maxsize = None

        self.method = method
        self.fill_value = fill_value
        self.maxsize = maxsize

        self.add_data(data)

    @property
    def method(self):
        """The method used for interpolating time."""
        return self._method

    @method.setter
    def method(self, val):
        if val not in TimeInterpolator.METHODS:
            raise ValueError("method not understood")
        self._method = val

    @property
    def fill_value(self):
        """Value to use when interpolating values outside of the data."""
        return self._fill_value

    @fill_value.setter
    def fill_value(self, val):
        if val != "extrapolate" and not isinstance(val, float):
            raise ValueError(
                "fill_value not understood, must be either float or 'extrapolate'"
            )
        self._fill_value = val

    @property
    def maxsize(self):
        """The maximum number of times to be stored."""
        return self._maxsize

    @maxsize.setter
    def maxsize(self, val):
        if val is not None and val <= _MINIMUM_SIZE_FOR_METHOD[self.method]:
            raise ValueError("maxsize too small for method ({0})".format(self.method))
        self._maxsize = val
        self._trim_data_to_maxsize()

    def add_data(self, time_and_data):
        """Add new data points to the interpolator.

        Parameters
        ----------
        time : float or iterable of float
            Time values
        data : float or iterable
            Data values to interpolate.
        """
        if isinstance(time_and_data, float):
            time_and_data = tuple(time_and_data)

        for t, d in time_and_data:
            self._insert_data(t, d)
        self._trim_data_to_maxsize()

    def _trim_data_to_maxsize(self):
        """Drop stored data to maxsize."""
        if self.maxsize is not None:
            while len(self._time) > self.maxsize:
                self._time.pop(0)
                self._data.pop(0)

    def _insert_data(self, time, data):
        """Insert data so that it is stored sorted by time."""
        ind = bisect.bisect_right(self._time, time)
        self._time.insert(ind, time)
        self._data.insert(ind, data)

        self._func = None

    def interpolate(self, time):
        """Interpolate the data at a given time."""
        if self._func is None:
            self._func = interp1d(
                self._time,
                self._data,
                axis=0,
                kind=self._method,
                fill_value=self._fill_value,
                copy=False,
                assume_sorted=True,
                bounds_error=False,
            )

        return self._func(time)

    def __call__(self, time):
        """Interpolate the data at a given time."""
        return self.interpolate(time)



import os

import six
import yaml

from .component import Component


def get_exchange_item_mapping(items):
    """Construct a mapping for exchange items.

    Parameters
    ----------
    items : iterable
        List of exchange item names or (*dest*, *src*) tuples.

    Returns
    -------
    items : list
        List of (*dest*, *src*) tuples of name mappings.

    Examples
    --------
    >>> from pymt.component.model import get_exchange_item_mapping
    >>> get_exchange_item_mapping([('foo', 'bar'), 'baz'])
    [('foo', 'bar'), ('baz', 'baz')]
    >>> get_exchange_item_mapping([dict(source='bar',
    ...                                 destination='foo'), 'baz'])
    [('foo', 'bar'), ('baz', 'baz')]
    """
    mapping = []
    for item in items:
        if isinstance(item, six.string_types):
            item_map = (item, item)
        else:
            try:
                item_map = (item["destination"], item["source"])
            except TypeError:
                item_map = item
        if len(item_map) != 2:
            raise ValueError(item)
        mapping.append(item_map)
    return mapping


class Model(object):
    def __init__(self, components, driver=None, duration=0.0):
        self._components = dict(components)
        self._driver = driver
        self._duration = duration

    def __getitem__(self, name):
        return self._components[name]

    @property
    def components(self):
        """Names of the components."""
        return list(self._components.keys())

    @property
    def driver(self):
        """Name of the driver component."""
        return self._driver

    @driver.setter
    def driver(self, driver):
        """Set the driver component."""
        if driver in self.components:
            self._driver = driver
        else:
            raise ValueError("%s not a component of the model" % driver)

    @property
    def duration(self):
        """Length of time the model will run."""
        return self._duration

    @duration.setter
    def duration(self, duration):
        """Set length of time the model will run."""
        self._duration = duration

    def go(self, filename=None):
        """Start the model."""
        if filename:
            with open(filename, "r") as f:
                model = yaml.safe_load(f.read())
            self._driver, self._duration = (model["driver"], model["duration"])

        self._components[self.driver].go(self.duration)

    @classmethod
    def load(cls, source):
        """Construct a model from a YAML-formatted string.

        Parameters
        ----------
        source : str or file_like
            YAML-formatted model configuration.

        Returns
        -------
        model : Model
            A newly-created Model.

        Examples
        --------
        >>> from pymt.framework.services import del_component_instances
        >>> del_component_instances(['air_port'])

        >>> from pymt.component.model import Model
        >>> source = '''
        ... name: air_port
        ... class: AirPort
        ... connectivity: []
        ... '''
        >>> model = Model.load(source)
        >>> model.components
        ['air_port']
        >>> model.driver is None
        True
        >>> model.duration
        0.0

        >>> model.driver = 'air_port'
        >>> model.duration = 1.
        >>> model['air_port'].current_time
        0.0
        >>> model.go() # doctest: +NORMALIZE_WHITESPACE
        {name: air_port, status: running, time: 1.0}
        >>> model['air_port'].current_time
        1.0
        """
        components, connectivities = Model.load_components(
            source, with_connectivities=True
        )

        for (name, component) in components.items():
            for port in connectivities[name]:
                mapping = get_exchange_item_mapping(port["exchange_items"])
                component.connect(
                    port["name"], components[port["connect"]], vars_to_map=mapping
                )

        return cls(components)

    @classmethod
    def from_file(cls, filename):
        """Construct a Model from the contents of a file.

        Parameters
        ----------
        filename : str
            Name of the model configuration file.

        Returns
        -------
        model : Model
            A newly-created model.
        """
        with open(filename, "r") as fp:
            return cls.load(fp)

    @classmethod
    def from_file_like(cls, file_like):
        """Construct a Model from the contents of a file-like object.

        Parameters
        ----------
        file_like : file_like
            A file-like object that contains the model configuration.

        Returns
        -------
        model : Model
            A newly-created model.
        """
        return cls.load(file_like)

    @classmethod
    def from_string(cls, contents):
        """Construct a Model from a string.

        Parameters
        ----------
        contents : str
            A string that contains the model configuration.

        Returns
        -------
        model : Model
            A newly-created model.
        """
        return cls.load(contents)

    @staticmethod
    def load_components(source, with_connectivities=False):
        """Construct a list of model components from a string.

        Parameters
        ----------
        source : str or file_like
            YAML-formatted components configuration.
        with_connectivities : boolean (optional)
            Return connectivity dictionary along with components.

        Returns
        -------
        components : dict
            Dictionary of newly-created components.

        Examples
        --------
        >>> from pymt.framework.services import del_component_instances
        >>> del_component_instances(['air_port'])

        >>> from pymt.component.model import Model
        >>> source = '''
        ... name: air_port
        ... class: AirPort
        ... connectivity: []
        ... '''
        >>> components = Model.load_components(source)
        >>> list(components.keys())
        ['air_port']
        """
        components = {}
        connectivities = {}

        prefix = os.getcwd()
        for section in yaml.safe_load_all(source):
            try:
                section["run_dir"] = os.path.join(prefix, section.get("run_dir", "."))
            except AttributeError:
                raise ValueError(section)

            components[section["name"]] = Component.from_dict(section)
            if with_connectivities:
                connectivities[section["name"]] = section["connectivity"]

        if with_connectivities:
            return components, connectivities
        else:
            return components
